@@ -27,24 +27,25 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 Conversion of markdown-formatted plain text to 'Pandoc' document.
 -}
-module Text.Pandoc.Readers.Markdown ( 
-                                     readMarkdown 
-                                    ) where
+module Text.Pandoc.Readers.Markdown ( readMarkdown ) where
 
-import Data.List ( transpose, isSuffixOf, sortBy, findIndex, intercalate )
+import Data.List ( transpose, sortBy, findIndex, intercalate )
+import qualified Data.Map as M
 import Data.Ord ( comparing )
 import Data.Char ( isAlphaNum )
 import Data.Maybe
 import Text.Pandoc.Definition
-import Text.Pandoc.Shared 
+import Text.Pandoc.Generic
+import Text.Pandoc.Shared
+import Text.Pandoc.Parsing
 import Text.Pandoc.Readers.LaTeX ( rawLaTeXInline, rawLaTeXEnvironment' )
-import Text.Pandoc.Readers.HTML ( rawHtmlBlock, anyHtmlBlockTag, 
-                                  anyHtmlInlineTag, anyHtmlTag,
-                                  anyHtmlEndTag, htmlEndTag, extractTagType,
-                                  htmlBlockElement, htmlComment, unsanitaryURI )
+import Text.Pandoc.Readers.HTML ( htmlTag, htmlInBalanced, isInlineTag, isBlockTag,
+                                  isTextTag, isCommentTag )
 import Text.Pandoc.CharacterReferences ( decodeCharacterReferences )
 import Text.ParserCombinators.Parsec
-import Control.Monad (when, liftM, unless)
+import Control.Monad (when, liftM, guard)
+import Text.HTML.TagSoup
+import Text.HTML.TagSoup.Match (tagOpen)
 
 -- | Read markdown from an input string and return a Pandoc document.
 readMarkdown :: ParserState -- ^ Parser state, including options for parser
@@ -56,18 +57,26 @@ readMarkdown state s = (readWith parseMarkdown) state (s ++ "\n\n")
 -- Constants and data structure definitions
 --
 
-bulletListMarkers :: [Char]
-bulletListMarkers = "*+-"
+isBulletListMarker :: Char -> Bool
+isBulletListMarker '*' = True
+isBulletListMarker '+' = True
+isBulletListMarker '-' = True
+isBulletListMarker _   = False
 
-hruleChars :: [Char]
-hruleChars = "*-_"
+isHruleChar :: Char -> Bool
+isHruleChar '*' = True
+isHruleChar '-' = True
+isHruleChar '_' = True
+isHruleChar _   = False
 
 setextHChars :: [Char]
 setextHChars = "=-"
 
--- treat these as potentially non-text when parsing inline:
-specialChars :: [Char]
-specialChars = "\\[]*_~`<>$!^-.&'\"\8216\8217\8220\8221;"
+isBlank :: Char -> Bool
+isBlank ' '  = True
+isBlank '\t' = True
+isBlank '\n' = True
+isBlank _    = False
 
 --
 -- auxiliary functions
@@ -104,12 +113,6 @@ failUnlessBeginningOfLine = do
   pos <- getPosition
   if sourceColumn pos == 1 then return () else fail "not beginning of line"
 
--- | Fail unless we're in "smart typography" mode.
-failUnlessSmart :: GenParser tok ParserState ()
-failUnlessSmart = do
-  state <- getState
-  if stateSmart state then return () else pzero
-
 -- | Parse a sequence of inline elements between square brackets,
 -- including inlines between balanced pairs of square brackets.
 inlinesInBalancedBrackets :: GenParser Char ParserState Inline
@@ -117,7 +120,7 @@ inlinesInBalancedBrackets :: GenParser Char ParserState Inline
 inlinesInBalancedBrackets parser = try $ do
   char '['
   result <- manyTill ( (do lookAhead $ try $ do (Str res) <- parser
-                                                unless (res == "[") pzero
+                                                guard (res == "[")
                            bal <- inlinesInBalancedBrackets parser
                            return $ [Str "["] ++ bal ++ [Str "]"])
                        <|> (count 1 parser))
@@ -141,7 +144,8 @@ authorsLine :: GenParser Char ParserState [[Inline]]
 authorsLine = try $ do 
   char '%'
   skipSpaces
-  authors <- sepEndBy (many (notFollowedBy (oneOf ";\n") >> inline))
+  authors <- sepEndBy (many (notFollowedBy (satisfy $ \c ->
+                                c == ';' || c == '\n') >> inline))
                        (char ';' <|>
                         try (newline >> notFollowedBy blankline >> spaceChar))
   newline
@@ -183,7 +187,18 @@ parseMarkdown = do
   -- now parse it for real...
   (title, author, date) <- option ([],[],[]) titleBlock
   blocks <- parseBlocks
-  return $ Pandoc (Meta title author date) $ filter (/= Null) blocks
+  let doc = Pandoc (Meta title author date) $ filter (/= Null) blocks
+  -- if there are labeled examples, change references into numbers
+  examples <- liftM stateExamples getState
+  let handleExampleRef :: Inline -> Inline
+      handleExampleRef z@(Str ('@':xs)) =
+        case M.lookup xs examples of
+              Just n     -> Str (show n)
+              Nothing    -> z
+      handleExampleRef z = z
+  if M.null examples
+     then return doc
+     else return $ bottomUp handleExampleRef doc
 
 -- 
 -- initial pass for references and notes
@@ -196,16 +211,24 @@ referenceKey = try $ do
   lab <- reference
   char ':'
   skipSpaces >> optional newline >> skipSpaces >> notFollowedBy (char '[')
-  let sourceURL excludes = many $
-        optional (char '\\') >> (noneOf (' ':excludes) <|> (notFollowedBy' referenceTitle >> char ' '))
-  src <- try (char '<' >> sourceURL ">\t\n" >>~ char '>') <|> sourceURL "\t\n"
+  let nl = char '\n' >> notFollowedBy blankline >> return ' '
+  let sourceURL = liftM unwords $ many $ try $ do
+                    notFollowedBy' referenceTitle
+                    skipMany spaceChar
+                    optional nl
+                    skipMany spaceChar
+                    notFollowedBy' reference
+                    many1 (satisfy $ not . isBlank)
+  let betweenAngles = try $ char '<' >>
+                            manyTill (noneOf ">\n" <|> nl) (char '>')
+  src <- try betweenAngles <|> sourceURL
   tit <- option "" referenceTitle
   blanklines
   endPos <- getPosition
-  let newkey = (lab, (escapeURI $ removeTrailingSpace src,  tit))
+  let target = (escapeURI $ removeTrailingSpace src,  tit)
   st <- getState
   let oldkeys = stateKeys st
-  updateState $ \s -> s { stateKeys = newkey : oldkeys }
+  updateState $ \s -> s { stateKeys = M.insert (toKey lab) target oldkeys }
   -- return blanks so line count isn't affected
   return $ replicate (sourceLine endPos - sourceLine startPos) '\n'
 
@@ -219,27 +242,32 @@ referenceTitle = try $ do
   return $ decodeCharacterReferences tit
 
 noteMarker :: GenParser Char ParserState [Char]
-noteMarker = skipNonindentSpaces >> string "[^" >> manyTill (noneOf " \t\n") (char ']')
+noteMarker = string "[^" >> many1Till (satisfy $ not . isBlank) (char ']')
 
 rawLine :: GenParser Char ParserState [Char]
-rawLine = do
+rawLine = try $ do
   notFollowedBy blankline
-  notFollowedBy' noteMarker
-  contents <- many1 nonEndline
-  end <- option "" (newline >> optional indentSpaces >> return "\n") 
-  return $ contents ++ end
+  notFollowedBy' $ try $ skipNonindentSpaces >> noteMarker
+  optional indentSpaces
+  anyLine
 
 rawLines :: GenParser Char ParserState [Char]
-rawLines = many1 rawLine >>= return . concat
+rawLines = do
+  first <- anyLine
+  rest <- many rawLine
+  return $ unlines (first:rest)
 
 noteBlock :: GenParser Char ParserState [Char]
 noteBlock = try $ do
   startPos <- getPosition
+  skipNonindentSpaces
   ref <- noteMarker
   char ':'
   optional blankline
   optional indentSpaces
-  raw <- sepBy rawLines (try (blankline >> indentSpaces))
+  raw <- sepBy rawLines
+             (try (blankline >> indentSpaces >>
+                   notFollowedBy blankline))
   optional blanklines
   endPos <- getPosition
   let newnote = (ref, (intercalate "\n" raw) ++ "\n\n")
@@ -271,6 +299,7 @@ block = do
                    , plain
                    , nullBlock ]
               else [ codeBlockDelimited
+                   , macro
                    , header 
                    , table
                    , codeBlockIndented
@@ -280,6 +309,7 @@ block = do
                    , bulletList
                    , orderedList
                    , definitionList
+                   , rawTeXBlock
                    , para
                    , rawHtmlBlocks
                    , plain
@@ -305,6 +335,9 @@ atxClosing = try $ skipMany (char '#') >> blanklines
 
 setextHeader :: GenParser Char ParserState Block
 setextHeader = try $ do
+  -- This lookahead prevents us from wasting time parsing Inlines
+  -- unless necessary -- it gives a significant performance boost.
+  lookAhead $ anyLine >> many1 (oneOf setextHChars) >> blankline
   text <- many1Till inline newline
   underlineChar <- oneOf setextHChars
   many (char underlineChar)
@@ -319,7 +352,7 @@ setextHeader = try $ do
 hrule :: GenParser Char st Block
 hrule = try $ do
   skipSpaces
-  start <- oneOf hruleChars
+  start <- satisfy isHruleChar
   count 2 (skipSpaces >> char start)
   skipMany (spaceChar <|> char start)
   newline
@@ -358,6 +391,7 @@ attributes = try $ do
 attribute :: GenParser Char st ([Char], [[Char]], [([Char], [Char])])
 attribute = identifierAttr <|> classAttr <|> keyValAttr
 
+
 identifier :: GenParser Char st [Char]
 identifier = do
   first <- letter
@@ -381,7 +415,7 @@ keyValAttr = try $ do
   key <- identifier
   char '='
   char '"'
-  val <- manyTill (noneOf "\n") (char '"')
+  val <- manyTill (satisfy (/='\n')) (char '"')
   return ("",[],[(key,val)])
 
 codeBlockDelimited :: GenParser Char st Block
@@ -476,7 +510,7 @@ bulletListStart = try $ do
   optional newline -- if preceded by a Plain block in a list context
   skipNonindentSpaces
   notFollowedBy' hrule     -- because hrules start out just like lists
-  oneOf bulletListMarkers
+  satisfy isBulletListMarker
   spaceChar
   skipSpaces
 
@@ -511,7 +545,7 @@ listLine = try $ do
   notFollowedBy' (do indentSpaces
                      many (spaceChar)
                      listStart)
-  chunks <- manyTill (htmlComment <|> count 1 anyChar) newline
+  chunks <- manyTill (liftM snd (htmlTag isCommentTag) <|> count 1 anyChar) newline
   return $ concat chunks ++ "\n"
 
 -- parse raw text for one list item, excluding start marker and continuations
@@ -631,23 +665,21 @@ definitionList = do
 --
 
 isHtmlOrBlank :: Inline -> Bool
-isHtmlOrBlank (HtmlInline _) = True
-isHtmlOrBlank (Space)        = True
-isHtmlOrBlank (LineBreak)    = True
-isHtmlOrBlank _              = False
+isHtmlOrBlank (RawInline "html" _) = True
+isHtmlOrBlank (Space)         = True
+isHtmlOrBlank (LineBreak)     = True
+isHtmlOrBlank _               = False
 
 para :: GenParser Char ParserState Block
 para = try $ do 
-  result <- many1 inline
-  if all isHtmlOrBlank result
-     then fail "treat as raw HTML"
-     else return ()
-  newline
-  blanklines <|> do st <- getState
-                    if stateStrict st
-                       then lookAhead (blockQuote <|> header) >> return ""
-                       else pzero
-  return $ Para $ normalizeSpaces result
+  result <- liftM normalizeSpaces $ many1 inline
+  guard $ not . all isHtmlOrBlank $ result
+  option (Plain result) $ try $ do
+              newline
+              blanklines <|>
+                (getState >>= guard . stateStrict >>
+                 lookAhead (blockQuote <|> header) >> return "")
+              return $ Para result
 
 plain :: GenParser Char ParserState Block
 plain = many1 inline >>~ spaces >>= return . Plain . normalizeSpaces
@@ -657,7 +689,7 @@ plain = many1 inline >>~ spaces >>= return . Plain . normalizeSpaces
 --
 
 htmlElement :: GenParser Char ParserState [Char]
-htmlElement = strictHtmlBlock <|> htmlBlockElement <?> "html element"
+htmlElement = strictHtmlBlock <|> liftM snd (htmlTag isBlockTag)
 
 htmlBlock :: GenParser Char ParserState Block
 htmlBlock = try $ do
@@ -665,27 +697,33 @@ htmlBlock = try $ do
     first <- htmlElement
     finalSpace <- many spaceChar
     finalNewlines <- many newline
-    return $ RawHtml $ first ++ finalSpace ++ finalNewlines
-
--- True if tag is self-closing
-isSelfClosing :: [Char] -> Bool
-isSelfClosing tag = 
-  isSuffixOf "/>" $ filter (not . (`elem` " \n\t")) tag
+    return $ RawBlock "html" $ first ++ finalSpace ++ finalNewlines
 
 strictHtmlBlock :: GenParser Char ParserState [Char]
-strictHtmlBlock = try $ do
-  tag <- anyHtmlBlockTag 
-  let tag' = extractTagType tag
-  if isSelfClosing tag || tag' == "hr" 
-     then return tag
-     else do contents <- many (notFollowedBy' (htmlEndTag tag') >> 
-                               (htmlElement <|> (count 1 anyChar)))
-             end <- htmlEndTag tag'
-             return $ tag ++ concat contents ++ end
+strictHtmlBlock = do
+  failUnlessBeginningOfLine
+  htmlInBalanced (not . isInlineTag)
+
+rawVerbatimBlock :: GenParser Char ParserState String
+rawVerbatimBlock = try $ do
+  (TagOpen tag _, open) <- htmlTag (tagOpen (\t ->
+                                      t == "pre" || t == "style" || t == "script")
+                                     (const True))
+  contents <- manyTill anyChar (htmlTag (~== TagClose tag))
+  return $ open ++ contents ++ renderTags [TagClose tag]
+
+rawTeXBlock :: GenParser Char ParserState Block
+rawTeXBlock = do
+  failIfStrict
+  result <- liftM (RawBlock "latex") rawLaTeXEnvironment'
+          <|> liftM (RawBlock "context") rawConTeXtEnvironment'
+  spaces
+  return result
 
 rawHtmlBlocks :: GenParser Char ParserState Block
 rawHtmlBlocks = do
-  htmlBlocks <- many1 $ do (RawHtml blk) <- rawHtmlBlock
+  htmlBlocks <- many1 $ do blk <- rawVerbatimBlock <|>
+                                   liftM snd (htmlTag isBlockTag)
                            sps <- do sp1 <- many spaceChar
                                      sp2 <- option "" (blankline >> return "\n")
                                      sp3 <- many spaceChar
@@ -697,7 +735,7 @@ rawHtmlBlocks = do
                            return $ blk ++ sps
   let combined = concat htmlBlocks
   let combined' = if last combined == '\n' then init combined else combined
-  return $ RawHtml combined'
+  return $ RawBlock "html" combined'
 
 --
 -- Tables
@@ -715,7 +753,7 @@ dashedLine ch = do
 -- Parse a table header with dashed lines of '-' preceded by 
 -- one (or zero) line of text.
 simpleTableHeader :: Bool  -- ^ Headerless table 
-                  -> GenParser Char ParserState ([[Char]], [Alignment], [Int])
+                  -> GenParser Char ParserState ([[Block]], [Alignment], [Int])
 simpleTableHeader headless = try $ do
   rawContent  <- if headless
                     then return ""
@@ -734,7 +772,9 @@ simpleTableHeader headless = try $ do
   let rawHeads' = if headless
                      then replicate (length dashes) ""
                      else rawHeads 
-  return (rawHeads', aligns, indices)
+  heads <- mapM (parseFromString (many plain)) $
+             map removeLeadingTrailingSpace rawHeads'
+  return (heads, aligns, indices)
 
 -- Parse a table footer - dashed lines followed by blank line.
 tableFooter :: GenParser Char ParserState [Char]
@@ -763,65 +803,27 @@ multilineRow :: [Int]
              -> GenParser Char ParserState [[Block]]
 multilineRow indices = do
   colLines <- many1 (rawTableLine indices)
-  optional blanklines
   let cols = map unlines $ transpose colLines
   mapM (parseFromString (many plain)) cols
 
--- Calculate relative widths of table columns, based on indices
-widthsFromIndices :: Int      -- Number of columns on terminal
-                  -> [Int]    -- Indices
-                  -> [Double] -- Fractional relative sizes of columns
-widthsFromIndices _ [] = []  
-widthsFromIndices numColumns indices = 
-  let lengths' = zipWith (-) indices (0:indices)
-      lengths  = reverse $
-                 case reverse lengths' of
-                      []       -> []
-                      [x]      -> [x]
-                      -- compensate for the fact that intercolumn
-                      -- spaces are counted in widths of all columns
-                      -- but the last...
-                      (x:y:zs) -> if x < y && y - x <= 2
-                                     then y:y:zs
-                                     else x:y:zs
-      totLength = sum lengths
-      quotient = if totLength > numColumns
-                   then fromIntegral totLength
-                   else fromIntegral numColumns
-      fracs = map (\l -> (fromIntegral l) / quotient) lengths in
-  tail fracs
-
 -- Parses a table caption:  inlines beginning with 'Table:'
 -- and followed by blank lines.
 tableCaption :: GenParser Char ParserState [Inline]
 tableCaption = try $ do
   skipNonindentSpaces
-  string "Table:"
+  string ":" <|> string "Table:"
   result <- many1 inline
   blanklines
   return $ normalizeSpaces result
 
--- Parse a table using 'headerParser', 'lineParser', and 'footerParser'.
-tableWith :: GenParser Char ParserState ([[Char]], [Alignment], [Int])
-          -> ([Int] -> GenParser Char ParserState [[Block]])
-          -> GenParser Char ParserState end
-          -> GenParser Char ParserState Block
-tableWith headerParser lineParser footerParser = try $ do
-    (rawHeads, aligns, indices) <- headerParser
-    lines' <- many1Till (lineParser indices) footerParser
-    caption <- option [] tableCaption
-    heads <- mapM (parseFromString (many plain)) rawHeads
-    state <- getState
-    let numColumns = stateColumns state
-    let widths = widthsFromIndices numColumns indices
-    return $ Table caption aligns widths heads lines'
-
 -- Parse a simple table with '---' header and one line per row.
 simpleTable :: Bool  -- ^ Headerless table
             -> GenParser Char ParserState Block
 simpleTable headless = do
   Table c a _w h l <- tableWith (simpleTableHeader headless) tableLine
-               (if headless then tableFooter else tableFooter <|> blanklines)
+              (return ())
+              (if headless then tableFooter else tableFooter <|> blanklines)
+              tableCaption
   -- Simple tables get 0s for relative column widths (i.e., use default)
   return $ Table c a (replicate (length a) 0) h l
 
@@ -832,10 +834,10 @@ simpleTable headless = do
 multilineTable :: Bool -- ^ Headerless table
                -> GenParser Char ParserState Block
 multilineTable headless =
-  tableWith (multilineTableHeader headless) multilineRow tableFooter
+  tableWith (multilineTableHeader headless) multilineRow blanklines tableFooter tableCaption
 
 multilineTableHeader :: Bool -- ^ Headerless table
-                     -> GenParser Char ParserState ([String], [Alignment], [Int])
+                     -> GenParser Char ParserState ([[Block]], [Alignment], [Int])
 multilineTableHeader headless = try $ do
   if headless
      then return '\n'
@@ -859,7 +861,9 @@ multilineTableHeader headless = try $ do
   let rawHeads = if headless
                     then replicate (length dashes) ""
                     else map (intercalate " ") rawHeadsList
-  return ((map removeLeadingTrailingSpace rawHeads), aligns, indices)
+  heads <- mapM (parseFromString (many plain)) $
+             map removeLeadingTrailingSpace rawHeads
+  return (heads, aligns, indices)
 
 -- Returns an alignment type for a table, based on a list of strings
 -- (the rows of the column header) and a number (the length of the
@@ -869,19 +873,25 @@ alignType :: [String]
           -> Alignment
 alignType [] _ = AlignDefault
 alignType strLst len =
-  let s          = head $ sortBy (comparing length) $ 
-                          map removeTrailingSpace strLst
-      leftSpace  = if null s then False else (s !! 0) `elem` " \t"
-      rightSpace = length s < len || (s !! (len - 1)) `elem` " \t"
+  let nonempties = filter (not . null) $ map removeTrailingSpace strLst
+      (leftSpace, rightSpace) =
+           case sortBy (comparing length) nonempties of
+                 (x:_)  -> (head x `elem` " \t", length x < len)
+                 []     -> (False, False)
   in  case (leftSpace, rightSpace) of
         (True,  False)   -> AlignRight
         (False, True)    -> AlignLeft
         (True,  True)    -> AlignCenter
         (False, False)   -> AlignDefault
 
+gridTable :: Bool -- ^ Headerless table
+          -> GenParser Char ParserState Block
+gridTable = gridTableWith block tableCaption
+
 table :: GenParser Char ParserState Block
 table = multilineTable False <|> simpleTable True <|>
-        simpleTable False <|> multilineTable True <?> "table"
+        simpleTable False <|> multilineTable True <|>
+        gridTable False <|> gridTable True <?> "table"
 
 -- 
 -- inline
@@ -891,30 +901,29 @@ inline :: GenParser Char ParserState Inline
 inline = choice inlineParsers <?> "inline"
 
 inlineParsers :: [GenParser Char ParserState Inline]
-inlineParsers = [ str
-                , smartPunctuation
-                , whitespace
+inlineParsers = [ whitespace
+                , str
                 , endline
                 , code
-                , charRef
                 , (fourOrMore '*' <|> fourOrMore '_')
                 , strong
                 , emph
                 , note
-                , inlineNote
                 , link
-#ifdef _CITEPROC
-                , inlineCitation
-#endif
+                , cite
                 , image
                 , math
                 , strikeout
                 , superscript
                 , subscript
+                , inlineNote  -- after superscript because of ^[link](/foo)^
                 , autoLink
-                , rawHtmlInline'
+                , rawHtmlInline
                 , rawLaTeXInline'
                 , escapedChar
+                , exampleRef
+                , smartPunctuation inline
+                , charRef
                 , symbol
                 , ltSign ]
 
@@ -928,12 +937,12 @@ failIfLink (Link _ _) = pzero
 failIfLink elt        = return elt
 
 escapedChar :: GenParser Char ParserState Inline
-escapedChar = do
+escapedChar = try $ do
   char '\\'
   state <- getState
-  result <- option '\\' $ if stateStrict state 
-                             then oneOf "\\`*_{}[]()>#+-.!~"
-                             else satisfy (not . isAlphaNum)
+  result <- if stateStrict state 
+               then oneOf "\\`*_{}[]()>#+-.!~"
+               else satisfy (not . isAlphaNum)
   return $ case result of
                 ' '   -> Str "\160" -- "\ " is a nonbreaking space
                 '\n'  -> LineBreak  -- "\[newline]" is a linebreak
@@ -947,12 +956,21 @@ ltSign = do
      else notFollowedBy' rawHtmlBlocks >> char '<' -- unless it starts html
   return $ Str ['<']
 
-specialCharsMinusLt :: [Char]
-specialCharsMinusLt = filter (/= '<') specialChars
+exampleRef :: GenParser Char ParserState Inline
+exampleRef = try $ do
+  char '@'
+  lab <- many1 (alphaNum <|> oneOf "-_")
+  -- We just return a Str. These are replaced with numbers
+  -- later. See the end of parseMarkdown.
+  return $ Str $ '@' : lab
 
 symbol :: GenParser Char ParserState Inline
 symbol = do 
-  result <- oneOf specialCharsMinusLt
+  result <- noneOf "<\\\n\t "
+         <|> try (do lookAhead $ char '\\'
+                     notFollowedBy' $ rawLaTeXEnvironment'
+                                   <|> rawConTeXtEnvironment'
+                     char '\\')
   return $ Str [result]
 
 -- parses inline code, between n `s and n `s
@@ -964,7 +982,8 @@ code = try $ do
                        (char '\n' >> notFollowedBy' blankline >> return " "))
                       (try (skipSpaces >> count (length starts) (char '`') >> 
                       notFollowedBy (char '`')))
-  return $ Code $ removeLeadingTrailingSpace $ concat result
+  attr <- option ([],[],[]) (try $ optional whitespace >> attributes)
+  return $ Code attr $ removeLeadingTrailingSpace $ concat result
 
 mathWord :: GenParser Char st [Char]
 mathWord = liftM concat $ many1 mathChunk
@@ -973,11 +992,11 @@ mathChunk :: GenParser Char st [Char]
 mathChunk = do char '\\'
                c <- anyChar
                return ['\\',c]
-        <|> many1 (noneOf " \t\n\\$")
+        <|> many1 (satisfy $ \c -> not (isBlank c || c == '\\' || c == '$'))
 
 math :: GenParser Char ParserState Inline
-math = (mathDisplay >>= return . Math DisplayMath)
-     <|> (mathInline >>= return . Math InlineMath)
+math =  (mathDisplay >>= applyMacros' >>= return . Math DisplayMath)
+     <|> (mathInline >>= applyMacros' >>= return . Math InlineMath)
 
 mathDisplay :: GenParser Char ParserState String 
 mathDisplay = try $ do
@@ -1026,87 +1045,6 @@ subscript = failIfStrict >> enclosed (char '~') (char '~')
             (notFollowedBy spaceChar >> inline) >>=  -- may not contain Space
             return . Subscript 
 
-smartPunctuation :: GenParser Char ParserState Inline
-smartPunctuation = failUnlessSmart >> 
-                   choice [ quoted, apostrophe, dash, ellipses ]
-
-apostrophe :: GenParser Char ParserState Inline
-apostrophe = (char '\'' <|> char '\8217') >> return Apostrophe
-
-quoted :: GenParser Char ParserState Inline
-quoted = doubleQuoted <|> singleQuoted 
-
-withQuoteContext :: QuoteContext
-                 -> (GenParser Char ParserState Inline)
-                 -> GenParser Char ParserState Inline
-withQuoteContext context parser = do
-  oldState <- getState
-  let oldQuoteContext = stateQuoteContext oldState
-  setState oldState { stateQuoteContext = context }
-  result <- parser
-  newState <- getState
-  setState newState { stateQuoteContext = oldQuoteContext }
-  return result
-
-singleQuoted :: GenParser Char ParserState Inline
-singleQuoted = try $ do
-  singleQuoteStart
-  withQuoteContext InSingleQuote $ many1Till inline singleQuoteEnd >>=
-    return . Quoted SingleQuote . normalizeSpaces
-
-doubleQuoted :: GenParser Char ParserState Inline
-doubleQuoted = try $ do 
-  doubleQuoteStart
-  withQuoteContext InDoubleQuote $ many1Till inline doubleQuoteEnd >>=
-    return . Quoted DoubleQuote . normalizeSpaces
-
-failIfInQuoteContext :: QuoteContext -> GenParser tok ParserState ()
-failIfInQuoteContext context = do
-  st <- getState
-  if stateQuoteContext st == context
-     then fail "already inside quotes"
-     else return ()
-
-singleQuoteStart :: GenParser Char ParserState Char
-singleQuoteStart = do 
-  failIfInQuoteContext InSingleQuote
-  char '\8216' <|> 
-     (try $ do char '\''  
-               notFollowedBy (oneOf ")!],.;:-? \t\n")
-               notFollowedBy (try (oneOfStrings ["s","t","m","ve","ll","re"] >>
-                                   satisfy (not . isAlphaNum))) 
-                                   -- possess/contraction
-               return '\'')
-
-singleQuoteEnd :: GenParser Char st Char
-singleQuoteEnd = try $ do
-  char '\8217' <|> char '\''
-  notFollowedBy alphaNum
-  return '\''
-
-doubleQuoteStart :: GenParser Char ParserState Char
-doubleQuoteStart = do
-  failIfInQuoteContext InDoubleQuote
-  char '\8220' <|>
-     (try $ do char '"'
-               notFollowedBy (oneOf " \t\n")
-               return '"')
-
-doubleQuoteEnd :: GenParser Char st Char
-doubleQuoteEnd = char '\8221' <|> char '"'
-
-ellipses :: GenParser Char st Inline
-ellipses = oneOfStrings ["...", " . . . ", ". . .", " . . ."] >> return Ellipses
-
-dash :: GenParser Char st Inline
-dash = enDash <|> emDash
-
-enDash :: GenParser Char st Inline
-enDash = try $ char '-' >> notFollowedBy (noneOf "0123456789") >> return EnDash
-
-emDash :: GenParser Char st Inline
-emDash = oneOfStrings ["---", "--"] >> return EmDash
-
 whitespace :: GenParser Char ParserState Inline
 whitespace = spaceChar >>
   (   (spaceChar >> skipMany spaceChar >> option Space (endline >> return LineBreak))
@@ -1115,20 +1053,19 @@ whitespace = spaceChar >>
 nonEndline :: GenParser Char st Char
 nonEndline = satisfy (/='\n')
 
-strChar :: GenParser Char st Char
-strChar = noneOf (specialChars ++ " \t\n")
-
 str :: GenParser Char ParserState Inline
 str = do
-  result <- many1 strChar
+  a <- alphaNum
+  as <- many $ alphaNum <|> (try $ char '_' >>~ lookAhead alphaNum)
+  let result = a:as
   state <- getState
   let spacesToNbr = map (\c -> if c == ' ' then '\160' else c)
   if stateSmart state
      then case likelyAbbrev result of
                []        -> return $ Str result
                xs        -> choice (map (\x ->
-                               try (string x >> char ' ' >>
-                                    notFollowedBy spaceChar >>
+                               try (string x >> oneOf " \n" >>
+                                    lookAhead alphaNum >>
                                     return (Str $ result ++ spacesToNbr x ++ "\160"))) xs)
                            <|> (return $ Str result)
      else return $ Str result
@@ -1151,15 +1088,13 @@ endline = try $ do
   newline
   notFollowedBy blankline
   st <- getState
-  if stateStrict st 
-    then do notFollowedBy emailBlockQuoteStart
-            notFollowedBy (char '#')  -- atx header
-    else return () 
+  when (stateStrict st) $ do
+    notFollowedBy emailBlockQuoteStart
+    notFollowedBy (char '#')  -- atx header
   -- parse potential list-starts differently if in a list:
-  if stateParserContext st == ListItemState
-     then notFollowedBy' (bulletListStart <|> 
-                          (anyOrderedListStart >> return ()))
-     else return ()
+  when (stateParserContext st == ListItemState) $ do
+     notFollowedBy' bulletListStart
+     notFollowedBy' anyOrderedListStart
   return Space
 
 --
@@ -1184,9 +1119,16 @@ source =
 source' :: GenParser Char st (String, [Char])
 source' = do
   skipSpaces
-  let sourceURL excludes = many $
-        optional (char '\\') >> (noneOf (' ':excludes) <|> (notFollowedBy' linkTitle >> char ' '))
-  src <- try (char '<' >> sourceURL ">\t\n" >>~ char '>') <|> sourceURL "\t\n"
+  let nl = char '\n' >>~ notFollowedBy blankline
+  let sourceURL = liftM unwords $ many $ try $ do
+                    notFollowedBy' linkTitle
+                    skipMany spaceChar
+                    optional nl
+                    skipMany spaceChar
+                    many1 (satisfy $ not . isBlank)
+  let betweenAngles = try $ char '<' >>
+                            manyTill (noneOf ">\n" <|> nl) (char '>')
+  src <- try betweenAngles <|> sourceURL
   tit <- option "" linkTitle
   skipSpaces
   eof
@@ -1205,10 +1147,7 @@ link :: GenParser Char ParserState Inline
 link = try $ do
   lab <- reference
   (src, tit) <- source <|> referenceLink lab
-  sanitize <- getState >>= return . stateSanitizeHTML
-  if sanitize && unsanitaryURI src
-     then fail "Unsanitary URI"
-     else return $ Link lab (src, tit)
+  return $ Link lab (src, tit)
 
 -- a link like [this][ref] or [this][] or [this]
 referenceLink :: [Inline]
@@ -1218,7 +1157,7 @@ referenceLink lab = do
                          optional (newline >> skipSpaces) >> reference))
   let ref' = if null ref then lab else ref
   state <- getState
-  case lookupKeySrc (stateKeys state) ref' of
+  case lookupKeySrc (stateKeys state) (toKey ref') of
      Nothing     -> fail "no corresponding key" 
      Just target -> return target 
 
@@ -1228,12 +1167,9 @@ autoLink = try $ do
   (orig, src) <- uri <|> emailAddress
   char '>'
   st <- getState
-  let sanitize = stateSanitizeHTML st
-  if sanitize && unsanitaryURI src
-     then fail "Unsanitary URI"
-     else return $ if stateStrict st
-                      then Link [Str orig] (src, "")
-                      else Link [Code orig] (src, "")
+  return $ if stateStrict st
+              then Link [Str orig] (src, "")
+              else Link [Code ("",["url"],[]) orig] (src, "")
 
 image :: GenParser Char ParserState Inline
 image = try $ do
@@ -1259,11 +1195,13 @@ inlineNote = try $ do
   return $ Note [Para contents]
 
 rawLaTeXInline' :: GenParser Char ParserState Inline
-rawLaTeXInline' = do
+rawLaTeXInline' = try $ do
   failIfStrict
-  (rawConTeXtEnvironment' >>= return . TeX)
-    <|> (rawLaTeXEnvironment' >>= return . TeX)
-    <|> rawLaTeXInline
+  lookAhead $ char '\\'
+  notFollowedBy' $ rawLaTeXEnvironment'
+                <|> rawConTeXtEnvironment'
+  RawInline _ s <- rawLaTeXInline
+  return $ RawInline "tex" s  -- "tex" because it might be context or latex
 
 rawConTeXtEnvironment' :: GenParser Char st String
 rawConTeXtEnvironment' = try $ do
@@ -1281,46 +1219,98 @@ inBrackets parser = do
   char ']'
   return $ "[" ++ contents ++ "]"
 
-rawHtmlInline' :: GenParser Char ParserState Inline
-rawHtmlInline' = do
+rawHtmlInline :: GenParser Char ParserState Inline
+rawHtmlInline = do
   st <- getState
-  result <- if stateStrict st
-               then choice [htmlBlockElement, anyHtmlTag, anyHtmlEndTag] 
-               else anyHtmlInlineTag
-  return $ HtmlInline result
+  (_,result) <- if stateStrict st
+                   then htmlTag (not . isTextTag)
+                   else htmlTag isInlineTag
+  return $ RawInline "html" result
 
-#ifdef _CITEPROC
-inlineCitation :: GenParser Char ParserState Inline
-inlineCitation = try $ do
+-- Citations
+
+cite :: GenParser Char ParserState Inline
+cite = do
   failIfStrict
-  cit <- citeMarker
-  let citations = readWith parseCitation defaultParserState cit
-  mr <- mapM chkCit citations
-  if catMaybes mr /= []
-     then return $ Cite citations []
-     else fail "no citation found"
+  citations <- textualCite <|> normalCite
+  return $ Cite citations []
 
-chkCit :: Target -> GenParser Char ParserState (Maybe Target)
-chkCit t = do
+spnl :: GenParser Char st ()
+spnl = try $ do
+  skipSpaces
+  optional newline
+  skipSpaces
+  notFollowedBy (char '\n')
+
+textualCite :: GenParser Char ParserState [Citation]
+textualCite = try $ do
+  (_, key) <- citeKey
+  let first = Citation{ citationId      = key
+                      , citationPrefix  = []
+                      , citationSuffix  = []
+                      , citationMode    = AuthorInText
+                      , citationNoteNum = 0
+                      , citationHash    = 0
+                      }
+  rest <- option [] $ try $ spnl >> normalCite
+  if null rest
+     then option [first] $ bareloc first
+     else return $ first : rest
+
+bareloc :: Citation -> GenParser Char ParserState [Citation]
+bareloc c = try $ do
+  spnl
+  char '['
+  suff <- suffix
+  rest <- option [] $ try $ char ';' >> citeList
+  spnl
+  char ']'
+  return $ c{ citationSuffix = suff } : rest
+
+normalCite :: GenParser Char ParserState [Citation]
+normalCite = try $ do
+  char '['
+  spnl
+  citations <- citeList
+  spnl
+  char ']'
+  return citations
+
+citeKey :: GenParser Char ParserState (Bool, String)
+citeKey = try $ do
+  suppress_author <- option False (char '-' >> return True)
+  char '@'
+  first <- letter
+  rest <- many $ (noneOf ",;]@ \t\n")
+  let key = first:rest
   st <- getState
-  case lookupKeySrc (stateKeys st) [Str $ fst t] of
-     Just  _ -> fail "This is a link"
-     Nothing -> if elem (fst t) $ stateCitations st
-                   then return $ Just t
-                   else return $ Nothing
+  guard $ key `elem` stateCitations st
+  return (suppress_author, key)
 
-citeMarker :: GenParser Char ParserState String
-citeMarker = char '[' >> manyTill ( noneOf "\n" <|> (newline >>~ notFollowedBy blankline) ) (char ']')
+suffix :: GenParser Char ParserState [Inline]
+suffix = try $ do
+  spnl
+  liftM normalizeSpaces $ many $ notFollowedBy (oneOf ";]") >> inline
 
-parseCitation :: GenParser Char ParserState [(String,String)]
-parseCitation = try $ sepBy (parseLabel) (oneOf ";")
+prefix :: GenParser Char ParserState [Inline]
+prefix = liftM normalizeSpaces $
+  manyTill inline (char ']' <|> liftM (const ']') (lookAhead citeKey))
 
-parseLabel :: GenParser Char ParserState (String,String)
-parseLabel = try $ do
-  res <- sepBy (skipSpaces >> optional newline >> skipSpaces >> many1 (noneOf "@;")) (oneOf "@")
-  case res of
-    [lab,loc] -> return (lab, loc)
-    [lab]     -> return (lab, "" )
-    _         -> return ("" , "" )
+citeList :: GenParser Char ParserState [Citation]
+citeList = sepBy1 citation (try $ char ';' >> spnl)
+
+citation :: GenParser Char ParserState Citation
+citation = try $ do
+  pref <- prefix
+  (suppress_author, key) <- citeKey
+  suff <- suffix
+  return $ Citation{ citationId        = key
+                     , citationPrefix  = pref
+                     , citationSuffix  = suff
+                     , citationMode    = if suppress_author
+                                            then SuppressAuthor
+                                            else NormalCitation
+                     , citationNoteNum = 0
+                     , citationHash    = 0
+                     }
 
-#endif
