@@ -31,10 +31,14 @@ module Text.Pandoc.Readers.RST (
                                 readRST
                                ) where
 import Text.Pandoc.Definition
-import Text.Pandoc.Shared 
+import Text.Pandoc.Shared
+import Text.Pandoc.Parsing
 import Text.ParserCombinators.Parsec
-import Control.Monad ( when, unless, liftM )
-import Data.List ( findIndex, delete, intercalate, transpose )
+import Control.Monad ( when )
+import Data.List ( findIndex, intercalate, transpose, sort, deleteFirstsBy )
+import qualified Data.Map as M
+import Text.Printf ( printf )
+import Data.Maybe ( catMaybes )
 
 -- | Parse reStructuredText string and return Pandoc document.
 readRST :: ParserState -- ^ Parser state, including options for parser
@@ -54,7 +58,7 @@ underlineChars = "!\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~"
 
 -- treat these as potentially non-text when parsing inline:
 specialChars :: [Char]
-specialChars = "\\`|*_<>$:[-"
+specialChars = "\\`|*_<>$:[-.\"'\8216\8217\8220\8221"
 
 --
 -- parsing documents
@@ -87,15 +91,17 @@ titleTransform blocks = (blocks, [])
 
 parseRST :: GenParser Char ParserState Pandoc
 parseRST = do
+  optional blanklines -- skip blank lines at beginning of file
   startPos <- getPosition
-  -- go through once just to get list of reference keys
+  -- go through once just to get list of reference keys and notes
   -- docMinusKeys is the raw document with blanks where the keys were...
-  docMinusKeys <- manyTill (referenceKey <|> lineClump) eof >>= return . concat
+  docMinusKeys <- manyTill (referenceKey <|> noteBlock <|> lineClump) eof >>=
+                   return . concat
   setInput docMinusKeys
   setPosition startPos
-  st <- getState
-  let reversedKeys = stateKeys st
-  updateState $ \s -> s { stateKeys = reverse reversedKeys }
+  st' <- getState
+  let reversedNotes = stateNotes st'
+  updateState $ \s -> s { stateNotes = reverse reversedNotes }
   -- now parse it for real...
   blocks <- parseBlocks 
   let blocks' = filter (/= Null) blocks
@@ -117,10 +123,9 @@ parseBlocks = manyTill block eof
 
 block :: GenParser Char ParserState Block
 block = choice [ codeBlock
-               , rawHtmlBlock
-               , rawLaTeXBlock
-               , fieldList
+               , rawBlock
                , blockQuote
+               , fieldList
                , imageBlock
                , customCodeBlock
                , unknownDirective
@@ -138,46 +143,54 @@ block = choice [ codeBlock
 -- field list
 --
 
-fieldListItem :: String -> GenParser Char st ([Char], [Char])
-fieldListItem indent = try $ do
+rawFieldListItem :: String -> GenParser Char ParserState (String, String)
+rawFieldListItem indent = try $ do
   string indent
   char ':'
-  name <- many1 alphaNum
+  name <- many1 $ alphaNum <|> spaceChar
   string ": "
   skipSpaces
   first <- manyTill anyChar newline
-  rest <- option "" $ try $ lookAhead (string indent >> oneOf " \t") >> 
-                            indentedBlock
-  return (name, intercalate " " (first:(lines rest)))
+  rest <- option "" $ try $ do lookAhead (string indent >> spaceChar)
+                               indentedBlock
+  let raw = first ++ "\n" ++ rest ++ "\n"
+  return (name, raw)
+
+fieldListItem :: String
+              -> GenParser Char ParserState (Maybe ([Inline], [[Block]]))
+fieldListItem indent = try $ do
+  (name, raw) <- rawFieldListItem indent
+  let term = [Str name]
+  contents <- parseFromString (many block) raw
+  case (name, contents) of
+       ("Author", x) -> do
+           updateState $ \st ->
+             st{ stateAuthors = stateAuthors st ++ [extractContents x] }
+           return Nothing
+       ("Authors", [BulletList auths]) -> do
+           updateState $ \st -> st{ stateAuthors = map extractContents auths }
+           return Nothing
+       ("Date", x) -> do
+           updateState $ \st -> st{ stateDate = extractContents x }
+           return Nothing
+       ("Title", x) -> do
+           updateState $ \st -> st{ stateTitle = extractContents x }
+           return Nothing
+       _            -> return $ Just (term, [contents])
+
+extractContents :: [Block] -> [Inline]
+extractContents [Plain auth] = auth
+extractContents [Para auth]  = auth
+extractContents _            = []
 
 fieldList :: GenParser Char ParserState Block
 fieldList = try $ do
-  indent <- lookAhead $ many (oneOf " \t")
+  indent <- lookAhead $ many spaceChar
   items <- many1 $ fieldListItem indent
   blanklines
-  let authors = case lookup "Authors" items of
-                  Just auth -> [auth]
-                  Nothing  -> map snd (filter (\(x,_) -> x == "Author") items)
-  unless (null authors) $ do
-    authors' <- mapM (parseFromString (many inline)) authors
-    updateState $ \st -> st {stateAuthors = map normalizeSpaces authors'}
-  case (lookup "Date" items) of
-           Just dat -> do
-                  dat' <- parseFromString (many inline) dat
-                  updateState $ \st -> st{ stateDate = normalizeSpaces dat' }
-           Nothing  -> return ()
-  case (lookup "Title" items) of
-           Just tit -> parseFromString (many inline) tit >>=
-                       \t -> updateState $ \st -> st {stateTitle = t}
-           Nothing  -> return ()
-  let remaining = filter (\(x,_) -> (x /= "Authors") && (x /= "Author") && 
-                  (x /= "Date") && (x /= "Title")) items
-  if null remaining
-              then return Null
-              else do terms <- mapM (return . (:[]) . Str . fst) remaining
-                      defs  <- mapM (parseFromString (many block) . snd) 
-                                    remaining
-                      return $ DefinitionList $ zip terms $ map (:[]) defs
+  if null items
+     then return Null
+     else return $ DefinitionList $ catMaybes items
 
 --
 -- line block
@@ -186,7 +199,7 @@ fieldList = try $ do
 lineBlockLine :: GenParser Char ParserState [Inline]
 lineBlockLine = try $ do
   string "| "
-  white <- many (oneOf " \t")
+  white <- many spaceChar
   line <- many $ (notFollowedBy newline >> inline) <|> (try $ endline >>~ char ' ')
   optional endline
   return $ normalizeSpaces $ (if null white then [] else [Str white]) ++ line
@@ -231,15 +244,16 @@ plain = many1 inline >>= return . Plain . normalizeSpaces
 -- image block
 --
 
-imageBlock :: GenParser Char st Block
+imageBlock :: GenParser Char ParserState Block
 imageBlock = try $ do
   string ".. image:: "
   src <- manyTill anyChar newline
-  fields <- option [] $ do indent <- lookAhead $ many (oneOf " /t")
-                           many1 $ fieldListItem indent
+  fields <- try $ do indent <- lookAhead $ many (oneOf " /t")
+                     many $ rawFieldListItem indent
   optional blanklines
   case lookup "alt" fields of
-        Just alt -> return $ Plain [Image [Str alt] (src, alt)]
+        Just alt -> return $ Plain [Image [Str $ removeTrailingSpace alt]
+                             (src, "")]
         Nothing  -> return $ Plain [Image [Str "image"] (src, "")]
 --
 -- header blocks
@@ -314,20 +328,19 @@ hrule = try $ do
 indentedLine :: String -> GenParser Char st [Char]
 indentedLine indents = try $ do
   string indents
-  result <- manyTill anyChar newline
-  return $ result ++ "\n"
+  manyTill anyChar newline
 
 -- two or more indented lines, possibly separated by blank lines.
 -- any amount of indentation will work.
 indentedBlock :: GenParser Char st [Char]
-indentedBlock = do 
-  indents <- lookAhead $ many1 (oneOf " \t")
+indentedBlock = try $ do 
+  indents <- lookAhead $ many1 spaceChar
   lns <- many $ choice $ [ indentedLine indents,
                            try $ do b <- blanklines
                                     l <- indentedLine indents
                                     return (b ++ l) ]
-  optional blanklines 
-  return $ concat lns
+  optional blanklines
+  return $ unlines lns
 
 codeBlock :: GenParser Char st Block
 codeBlock = try $ do
@@ -365,23 +378,16 @@ birdTrackLine = do
   manyTill anyChar newline
 
 --
--- raw html
+-- raw html/latex/etc
 --
 
-rawHtmlBlock :: GenParser Char st Block
-rawHtmlBlock = try $ string ".. raw:: html" >> blanklines >>
-                     indentedBlock >>= return . RawHtml
-
---
--- raw latex
---
-
-rawLaTeXBlock :: GenParser Char st Block
-rawLaTeXBlock = try $ do
-  string ".. raw:: latex"
+rawBlock :: GenParser Char st Block
+rawBlock = try $ do
+  string ".. raw:: "
+  lang <- many1 (letter <|> digit)
   blanklines
   result <- indentedBlock
-  return $ Para [(TeX result)]
+  return $ RawBlock lang result
 
 --
 -- block quotes
@@ -408,7 +414,7 @@ definitionListItem = try $ do
   term <- many1Till inline endline
   raw <- indentedBlock
   -- parse the extracted block, which may contain various block elements:
-  contents <- parseFromString parseBlocks $ raw ++ "\n\n"
+  contents <- parseFromString parseBlocks $ raw ++ "\n"
   return (normalizeSpaces term, [contents])
 
 definitionList :: GenParser Char ParserState Block
@@ -425,7 +431,7 @@ bulletListStart = try $ do
 -- parses ordered list start and returns its length (inc following whitespace)
 orderedListStart :: ListNumberStyle
                  -> ListNumberDelim
-                 -> GenParser Char st Int
+                 -> GenParser Char ParserState Int
 orderedListStart style delim = try $ do
   (_, markerLen) <- withHorizDisplacement (orderedListMarker style delim)
   white <- many1 spaceChar
@@ -505,9 +511,35 @@ unknownDirective = try $ do
   string ".."
   notFollowedBy (noneOf " \t\n")
   manyTill anyChar newline
-  many $ blanklines <|> (oneOf " \t" >> manyTill anyChar newline)
+  many $ blanklines <|> (spaceChar >> manyTill anyChar newline)
   return Null
 
+---
+--- note block
+---
+
+noteBlock :: GenParser Char ParserState [Char]
+noteBlock = try $ do
+  startPos <- getPosition
+  string ".."
+  spaceChar >> skipMany spaceChar
+  ref <- noteMarker
+  spaceChar >> skipMany spaceChar
+  first <- anyLine
+  blanks <- option "" blanklines
+  rest <- option "" indentedBlock
+  endPos <- getPosition
+  let raw = first ++ "\n" ++ blanks ++ rest ++ "\n"
+  let newnote = (ref, raw)
+  st <- getState
+  let oldnotes = stateNotes st
+  updateState $ \s -> s { stateNotes = newnote : oldnotes }
+  -- return blanks so line count isn't affected
+  return $ replicate (sourceLine endPos - sourceLine startPos) '\n'
+
+noteMarker :: GenParser Char ParserState [Char]
+noteMarker = char '[' >> (many1 digit <|> count 1 (oneOf "#*")) >>~ char ']'
+
 -- 
 -- reference key
 --
@@ -540,10 +572,10 @@ referenceName = quotedReferenceName <|>
 referenceKey :: GenParser Char ParserState [Char]
 referenceKey = do
   startPos <- getPosition
-  key <- choice [imageKey, anonymousKey, regularKey]
+  (key, target) <- choice [imageKey, anonymousKey, regularKey]
   st <- getState
   let oldkeys = stateKeys st
-  updateState $ \s -> s { stateKeys = key : oldkeys }
+  updateState $ \s -> s { stateKeys = M.insert key target oldkeys }
   optional blanklines
   endPos <- getPosition
   -- return enough blanks to replace key
@@ -558,28 +590,29 @@ targetURI = do
   blanklines
   return $ escapeURI $ removeLeadingTrailingSpace $ contents
 
-imageKey :: GenParser Char ParserState ([Inline], (String, [Char]))
+imageKey :: GenParser Char ParserState (Key, Target)
 imageKey = try $ do
   string ".. |"
   ref <- manyTill inline (char '|')
   skipSpaces
   string "image::"
   src <- targetURI
-  return (normalizeSpaces ref, (src, ""))
+  return (toKey (normalizeSpaces ref), (src, ""))
 
-anonymousKey :: GenParser Char st ([Inline], (String, [Char]))
+anonymousKey :: GenParser Char st (Key, Target)
 anonymousKey = try $ do
   oneOfStrings [".. __:", "__"]
   src <- targetURI
-  return ([Str "_"], (src, ""))
+  pos <- getPosition
+  return (toKey [Str $ "_" ++ printf "%09d" (sourceLine pos)], (src, ""))
 
-regularKey :: GenParser Char ParserState ([Inline], (String, [Char]))
+regularKey :: GenParser Char ParserState (Key, Target)
 regularKey = try $ do
   string ".. _"
   ref <- referenceName
   char ':'
   src <- targetURI
-  return (normalizeSpaces ref, (src, ""))
+  return (toKey (normalizeSpaces ref), (src, ""))
 
 --
 -- tables
@@ -607,41 +640,20 @@ dashedLine ch = do
 simpleDashedLines :: Char -> GenParser Char st [(Int,Int)]
 simpleDashedLines ch = try $ many1 (dashedLine ch)
 
-gridPart :: Char -> GenParser Char st (Int, Int)
-gridPart ch = do
-  dashes <- many1 (char ch)
-  char '+'
-  return (length dashes, length dashes + 1)
-
-gridDashedLines :: Char -> GenParser Char st [(Int,Int)]
-gridDashedLines ch = try $ char '+' >> many1 (gridPart ch) >>~ blankline
-
 -- Parse a table row separator
 simpleTableSep :: Char -> GenParser Char ParserState Char
 simpleTableSep ch = try $ simpleDashedLines ch >> newline
 
-gridTableSep :: Char -> GenParser Char ParserState Char
-gridTableSep ch = try $ gridDashedLines ch >> return '\n'
-
 -- Parse a table footer
 simpleTableFooter :: GenParser Char ParserState [Char]
 simpleTableFooter = try $ simpleTableSep '=' >> blanklines
 
-gridTableFooter :: GenParser Char ParserState [Char]
-gridTableFooter = blanklines
-
 -- Parse a raw line and split it into chunks by indices.
 simpleTableRawLine :: [Int] -> GenParser Char ParserState [String]
 simpleTableRawLine indices = do
   line <- many1Till anyChar newline
   return (simpleTableSplitLine indices line)
 
-gridTableRawLine :: [Int] -> GenParser Char ParserState [String]
-gridTableRawLine indices = do
-  char '|'
-  line <- many1Till anyChar newline
-  return (gridTableSplitLine indices $ removeTrailingSpace line)
-
 -- Parse a table row and return a list of blocks (columns).
 simpleTableRow :: [Int] -> GenParser Char ParserState [[Block]]
 simpleTableRow indices = do
@@ -651,64 +663,13 @@ simpleTableRow indices = do
   let cols = map unlines . transpose $ firstLine : colLines
   mapM (parseFromString (many plain)) cols
 
-gridTableRow :: [Int]
-             -> GenParser Char ParserState [[Block]]
-gridTableRow indices = do
-  colLines <- many1 (gridTableRawLine indices)
-  let cols = map ((++ "\n") . unlines . removeOneLeadingSpace) $
-               transpose colLines
-  mapM (liftM compactifyCell . parseFromString (many block)) cols
-
-compactifyCell :: [Block] -> [Block]
-compactifyCell bs = head $ compactify [bs]
-
 simpleTableSplitLine :: [Int] -> String -> [String]
 simpleTableSplitLine indices line =
   map removeLeadingTrailingSpace
   $ tail $ splitByIndices (init indices) line
 
-gridTableSplitLine :: [Int] -> String -> [String]
-gridTableSplitLine indices line =
-  map removeFinalBar $ tail $ splitByIndices (init indices) line
-
-removeFinalBar :: String -> String
-removeFinalBar = reverse . dropWhile (=='|') .  dropWhile (`elem` " \t") .
-                 reverse
-
-removeOneLeadingSpace :: [String] -> [String]
-removeOneLeadingSpace xs =
-  if all startsWithSpace xs
-     then map (drop 1) xs
-     else xs
-   where startsWithSpace ""     = True
-         startsWithSpace (y:_) = y == ' '
-
--- Calculate relative widths of table columns, based on indices
-widthsFromIndices :: Int      -- Number of columns on terminal
-                  -> [Int]    -- Indices
-                  -> [Double] -- Fractional relative sizes of columns
-widthsFromIndices _ [] = []
-widthsFromIndices numColumns indices =
-  let lengths' = zipWith (-) indices (0:indices)
-      lengths  = reverse $
-                 case reverse lengths' of
-                      []       -> []
-                      [x]      -> [x]
-                      -- compensate for the fact that intercolumn
-                      -- spaces are counted in widths of all columns
-                      -- but the last...
-                      (x:y:zs) -> if x < y && y - x <= 2
-                                     then y:y:zs
-                                     else x:y:zs
-      totLength = sum lengths
-      quotient = if totLength > numColumns
-                   then fromIntegral totLength
-                   else fromIntegral numColumns
-      fracs = map (\l -> (fromIntegral l) / quotient) lengths in
-  tail fracs
-
 simpleTableHeader :: Bool  -- ^ Headerless table 
-                  -> GenParser Char ParserState ([[Char]], [Alignment], [Int])
+                  -> GenParser Char ParserState ([[Block]], [Alignment], [Int])
 simpleTableHeader headless = try $ do
   optional blanklines
   rawContent  <- if headless
@@ -722,64 +683,23 @@ simpleTableHeader headless = try $ do
   let rawHeads = if headless
                     then replicate (length dashes) ""
                     else simpleTableSplitLine indices rawContent
-  return (rawHeads, aligns, indices)
+  heads <- mapM (parseFromString (many plain)) $
+             map removeLeadingTrailingSpace rawHeads
+  return (heads, aligns, indices)
 
-gridTableHeader :: Bool -- ^ Headerless table
-                -> GenParser Char ParserState ([String], [Alignment], [Int])
-gridTableHeader headless = try $ do
-  optional blanklines
-  dashes <- gridDashedLines '-'
-  rawContent  <- if headless
-                    then return $ repeat "" 
-                    else many1
-                         (notFollowedBy (gridTableSep '=') >> char '|' >> many1Till anyChar newline)
-  if headless
-     then return ()
-     else gridTableSep '=' >> return ()
-  let lines'   = map snd dashes
-  let indices  = scanl (+) 0 lines'
-  let aligns   = replicate (length lines') AlignDefault -- RST does not have a notion of alignments
-  let rawHeads = if headless
-                    then replicate (length dashes) ""
-                    else map (intercalate " ") $ transpose
-                       $ map (gridTableSplitLine indices) rawContent
-  return (rawHeads, aligns, indices)
-
--- Parse a table using 'headerParser', 'lineParser', and 'footerParser'.
-tableWith :: GenParser Char ParserState ([[Char]], [Alignment], [Int])
-          -> ([Int] -> GenParser Char ParserState [[Block]])
-          -> GenParser Char ParserState sep
-          -> GenParser Char ParserState end
-          -> GenParser Char ParserState Block
-tableWith headerParser rowParser lineParser footerParser = try $ do
-    (rawHeads, aligns, indices) <- headerParser
-    lines' <- rowParser indices `sepEndBy` lineParser
-    footerParser
-    heads <- mapM (parseFromString (many plain)) rawHeads
-    state <- getState
-    let captions = [] -- no notion of captions in RST
-    let numColumns = stateColumns state
-    let widths = widthsFromIndices numColumns indices
-    return $ Table captions aligns widths heads lines'
-
--- Parse a simple table with '---' header and one line per row.
+-- Parse a simple table.
 simpleTable :: Bool  -- ^ Headerless table
             -> GenParser Char ParserState Block
 simpleTable headless = do
-  Table c a _w h l <- tableWith (simpleTableHeader headless) simpleTableRow sep simpleTableFooter
+  Table c a _w h l <- tableWith (simpleTableHeader headless) simpleTableRow sep simpleTableFooter (return [])
   -- Simple tables get 0s for relative column widths (i.e., use default)
   return $ Table c a (replicate (length a) 0) h l
  where
   sep = return () -- optional (simpleTableSep '-')
 
--- Parse a grid table:  starts with row of '-' on top, then header
--- (which may be grid), then the rows,
--- which may be grid, separated by blank lines, and
--- ending with a footer (dashed line followed by blank line).
 gridTable :: Bool -- ^ Headerless table
-               -> GenParser Char ParserState Block
-gridTable headless =
-  tableWith (gridTableHeader headless) gridTableRow (gridTableSep '-') gridTableFooter
+          -> GenParser Char ParserState Block
+gridTable = gridTableWith block (return [])
 
 table :: GenParser Char ParserState Block
 table = gridTable False <|> simpleTable False <|>
@@ -791,17 +711,19 @@ table = gridTable False <|> simpleTable False <|>
  --
 
 inline :: GenParser Char ParserState Inline
-inline = choice [ link
+inline = choice [ whitespace
+                , link
                 , str
-                , whitespace
                 , endline
                 , strong
                 , emph
                 , code
                 , image
-                , hyphens
                 , superscript
                 , subscript
+                , note
+                , smartPunctuation inline
+                , hyphens
                 , escapedChar
                 , symbol ] <?> "inline"
 
@@ -825,7 +747,8 @@ code :: GenParser Char ParserState Inline
 code = try $ do 
   string "``"
   result <- manyTill anyChar (try (string "``"))
-  return $ Code $ removeLeadingTrailingSpace $ intercalate " " $ lines result
+  return $ Code nullAttr
+         $ removeLeadingTrailingSpace $ intercalate " " $ lines result
 
 emph :: GenParser Char ParserState Inline
 emph = enclosed (char '*') (char '*') inline >>= 
@@ -889,17 +812,22 @@ explicitLink = try $ do
 referenceLink :: GenParser Char ParserState Inline
 referenceLink = try $ do
   label' <- (quotedReferenceName <|> simpleReferenceName) >>~ char '_'
-  key <- option label' (do{char '_'; return [Str "_"]}) -- anonymous link
   state <- getState
   let keyTable = stateKeys state
+  let isAnonKey x = case fromKey x of
+                         [Str ('_':_)] -> True
+                         _             -> False
+  key <- option (toKey label') $
+                do char '_'
+                   let anonKeys = sort $ filter isAnonKey $ M.keys keyTable
+                   if null anonKeys
+                      then pzero
+                      else return (head anonKeys)
   (src,tit) <- case lookupKeySrc keyTable key of
                     Nothing     -> fail "no corresponding key"
                     Just target -> return target
-  -- if anonymous link, remove first anon key so it won't be used again
-  let keyTable' = if (key == [Str "_"]) -- anonymous link? 
-                    then delete ([Str "_"], (src,tit)) keyTable -- remove first anon key 
-                    else keyTable                    
-  setState $ state { stateKeys = keyTable' }
+  -- if anonymous link, remove key so it won't be used again
+  when (isAnonKey key) $ updateState $ \s -> s{ stateKeys = M.delete key keyTable }
   return $ Link (normalizeSpaces label') (src, tit) 
 
 autoURI :: GenParser Char ParserState Inline
@@ -922,7 +850,24 @@ image = try $ do
   ref <- manyTill inline (char '|')
   state <- getState
   let keyTable = stateKeys state
-  (src,tit) <- case lookupKeySrc keyTable ref of
+  (src,tit) <- case lookupKeySrc keyTable (toKey ref) of
                      Nothing     -> fail "no corresponding key"
                      Just target -> return target
   return $ Image (normalizeSpaces ref) (src, tit)
+
+note :: GenParser Char ParserState Inline
+note = try $ do
+  ref <- noteMarker
+  char '_'
+  state <- getState
+  let notes = stateNotes state
+  case lookup ref notes of
+    Nothing   -> fail "note not found"
+    Just raw  -> do
+      contents <- parseFromString parseBlocks raw
+      when (ref == "*" || ref == "#") $ do -- auto-numbered
+        -- delete the note so the next auto-numbered note
+        -- doesn't get the same contents:
+        let newnotes = deleteFirstsBy (==) notes [(ref,raw)]
+        updateState $ \st -> st{ stateNotes = newnotes }
+      return $ Note contents
