@@ -72,7 +72,6 @@ import Text.ParserCombinators.Parsec
 import Control.Monad (liftM, when, forM)
 import System.FilePath
 import Data.List (intercalate, intersperse)
-import Text.PrettyPrint (text, Doc)
 import Text.XHtml (primHtml, Html)
 import Data.ByteString.Lazy.UTF8 (ByteString, fromString)
 import Text.Pandoc.Shared (readDataFile)
@@ -83,7 +82,6 @@ getDefaultTemplate :: (Maybe FilePath) -- ^ User data directory to search first
                    -> String           -- ^ Name of writer 
                    -> IO (Either E.IOException String)
 getDefaultTemplate _ "native" = return $ Right ""
-getDefaultTemplate user "s5" = getDefaultTemplate user "html"
 getDefaultTemplate user "odt" = getDefaultTemplate user "opendocument"
 getDefaultTemplate user writer = do
   let format = takeWhile (/='+') writer  -- strip off "+lhs" if present
@@ -113,9 +111,6 @@ instance TemplateTarget ByteString where
 instance TemplateTarget Html where
   toTarget = primHtml
 
-instance TemplateTarget Doc where
-  toTarget = text
-
 -- | Renders a template 
 renderTemplate :: TemplateTarget a
                => [(String,String)]  -- ^ Assoc. list of values for variables
@@ -173,7 +168,7 @@ for = try $ do
   string "$for("
   id' <- ident
   string ")$"
-  -- if newline after the "if", then a newline after "endif" will be swallowed
+  -- if newline after the "for", then a newline after "endfor" will be swallowed
   multiline <- option False $ try $ skipEndline >> return True
   let matches = filter (\(k,_) -> k == id') vars 
   let indent = replicate pos ' '
