@@ -34,14 +34,16 @@ import Text.Pandoc.CharacterReferences ( decodeCharacterReferences )
 import Text.Pandoc.Shared
 import Text.Pandoc.Templates
 import Text.Pandoc.Readers.TeXMath
-import Text.Pandoc.Highlighting ( highlightHtml )
+import Text.Pandoc.Highlighting ( highlightHtml, defaultHighlightingCss )
 import Text.Pandoc.XML (stripTags, escapeStringForXML)
+import Network.HTTP ( urlEncode )
 import Numeric ( showHex )
 import Data.Char ( ord, toLower )
 import Data.List ( isPrefixOf, intersperse )
 import Data.Maybe ( catMaybes )
 import Control.Monad.State
-import Text.XHtml.Transitional hiding ( stringToHtml )
+import Text.XHtml.Transitional hiding ( stringToHtml, unordList, ordList )
+import qualified Text.XHtml.Transitional as XHtml
 import Text.TeXMath
 import Text.XML.Light.Output
 
@@ -57,15 +59,18 @@ defaultWriterState = WriterState {stNotes= [], stMath = False, stHighlighting =
 
 -- Helpers to render HTML with the appropriate function.
 
-renderFragment :: (HTML html) => WriterOptions -> html -> String
-renderFragment opts = if writerWrapText opts
-                         then renderHtmlFragment
-                         else showHtmlFragment
-
 -- | Modified version of Text.XHtml's stringToHtml.
 -- Use unicode characters wherever possible.
-stringToHtml :: String -> Html
-stringToHtml = primHtml . escapeStringForXML
+stringToHtml :: WriterOptions -> String -> Html
+stringToHtml opts = if writerAscii opts
+                       then XHtml.stringToHtml
+                       else primHtml . escapeStringForXML
+
+-- | Hard linebreak.
+nl :: WriterOptions -> Html
+nl opts = if writerWrapText opts
+             then primHtml "\n"
+             else noHtml
 
 -- | Convert Pandoc document to Html string.
 writeHtmlString :: WriterOptions -> Pandoc -> String
@@ -74,7 +79,7 @@ writeHtmlString opts d =
                                                  defaultWriterState
   in  if writerStandalone opts
          then inTemplate opts tit auths date toc body' newvars
-         else renderFragment opts body'
+         else dropWhile (=='\n') $ showHtmlFragment body'
 
 -- | Convert Pandoc document to Html structure.
 writeHtml :: WriterOptions -> Pandoc -> Html
@@ -104,10 +109,27 @@ pandocToHtml opts (Pandoc (Meta title' authors' date') blocks) = do
   toc <- if writerTableOfContents opts 
             then tableOfContents opts sects
             else return Nothing
-  blocks' <- liftM toHtmlFromList $ mapM (elementToHtml opts) sects
+  let startSlide = RawBlock "html" "<div class=\"slide\">\n"
+      endSlide   = RawBlock "html" "</div>\n"
+  let cutUp (HorizontalRule : Header 1 ys : xs) = cutUp (Header 1 ys : xs)
+      cutUp (HorizontalRule : xs) = [endSlide, startSlide] ++ cutUp xs
+      cutUp (Header 1 ys : xs)    = [endSlide, startSlide] ++
+                                    (Header 1 ys : cutUp xs)
+      cutUp (x:xs)                = x : cutUp xs
+      cutUp []                    = [] 
+  let slides = case blocks of
+                (HorizontalRule : xs) -> [startSlide] ++ cutUp xs ++ [endSlide]
+                (Header 1 ys : xs)    -> [startSlide, Header 1 ys] ++
+                                           cutUp xs ++ [endSlide]
+                _                     -> [startSlide] ++ cutUp blocks ++
+                                           [endSlide]
+  blocks' <- liftM (toHtmlFromList . intersperse (nl opts)) $
+              if writerSlideVariant opts `elem` [SlidySlides, S5Slides]
+                 then mapM (blockToHtml opts) slides
+                 else mapM (elementToHtml opts) sects
   st <- get
   let notes = reverse (stNotes st)
-  let thebody = blocks' +++ footnoteSection notes
+  let thebody = blocks' +++ footnoteSection opts notes
   let  math = if stMath st
                 then case writerHTMLMathMethod opts of
                            LaTeXMathML (Just url) ->
@@ -116,6 +138,8 @@ pandocToHtml opts (Pandoc (Meta title' authors' date') blocks) = do
                            MathML (Just url) ->
                               script ! 
                               [src url, thetype "text/javascript"] $ noHtml
+                           MathJax url ->
+                              script ! [src url, thetype "text/javascript"] $ noHtml
                            JsMath (Just url) ->
                               script !
                               [src url, thetype "text/javascript"] $ noHtml
@@ -125,8 +149,9 @@ pandocToHtml opts (Pandoc (Meta title' authors' date') blocks) = do
                                            primHtml s
                                       Nothing -> noHtml
                 else noHtml
-  let newvars = [("highlighting","yes") | stHighlighting st] ++
-                [("math", renderHtmlFragment math) | stMath st]
+  let newvars = [("highlighting-css", defaultHighlightingCss) |
+                   stHighlighting st] ++
+                [("math", showHtmlFragment math) | stMath st]
   return (tit, auths, date, toc, thebody, newvars)
 
 inTemplate :: TemplateTarget a
@@ -145,12 +170,13 @@ inTemplate opts tit auths date toc body' newvars =
       date'       = stripTags $ showHtmlFragment date
       variables   = writerVariables opts ++ newvars
       context     = variables ++
-                    [ ("body", renderHtmlFragment body')
+                    [ ("body", dropWhile (=='\n') $ showHtmlFragment body')
                     , ("pagetitle", topTitle')
-                    , ("title", renderHtmlFragment tit)
+                    , ("title", dropWhile (=='\n') $ showHtmlFragment tit)
                     , ("date", date') ] ++
+                    [ ("html5","true") | writerHtml5 opts ] ++
                     (case toc of
-                         Just t  -> [ ("toc", renderHtmlFragment t)]
+                         Just t  -> [ ("toc", showHtmlFragment t)]
                          Nothing -> [])  ++
                     [ ("author", a) | a <- authors ]
   in  renderTemplate context $ writerTemplate opts
@@ -159,6 +185,14 @@ inTemplate opts tit auths date toc body' newvars =
 prefixedId :: WriterOptions -> String -> HtmlAttr
 prefixedId opts s = identifier $ writerIdentifierPrefix opts ++ s
 
+-- | Replacement for Text.XHtml's unordList.
+unordList :: WriterOptions -> ([Html] -> Html)
+unordList opts items = ulist << toListItems opts items
+
+-- | Replacement for Text.XHtml's ordList.
+ordList :: WriterOptions -> ([Html] -> Html)
+ordList opts items = olist << toListItems opts items
+
 -- | Construct table of contents from list of elements.
 tableOfContents :: WriterOptions -> [Element] -> State WriterState (Maybe Html)
 tableOfContents _ [] = return Nothing
@@ -168,7 +202,14 @@ tableOfContents opts sects = do
   let tocList = catMaybes contents
   return $ if null tocList
               then Nothing
-              else Just $ thediv ! [prefixedId opts' "TOC"] $ unordList tocList
+              else Just $
+                   if writerHtml5 opts
+                      then (tag "nav" ! [prefixedId opts' "TOC"] $
+                            nl opts +++ unordList opts tocList +++ nl opts)
+                            +++ nl opts
+                      else (thediv ! [prefixedId opts' "TOC"] $
+                            nl opts +++ unordList opts tocList +++ nl opts)
+                            +++ nl opts
 
 -- | Convert section number to string
 showSecNum :: [Int] -> String
@@ -181,13 +222,13 @@ elementToListItem _ (Blk _) = return Nothing
 elementToListItem opts (Sec _ num id' headerText subsecs) = do
   let sectnum = if writerNumberSections opts
                    then (thespan ! [theclass "toc-section-number"] << showSecNum num) +++
-                         stringToHtml " "
+                         stringToHtml opts" "
                    else noHtml
   txt <- liftM (sectnum +++) $ inlineListToHtml opts headerText
   subHeads <- mapM (elementToListItem opts) subsecs >>= return . catMaybes
   let subList = if null subHeads
                    then noHtml
-                   else unordList subHeads
+                   else unordList opts subHeads
   return $ Just $ (anchor ! [href ("#" ++ writerIdentifierPrefix opts ++ id')] $ txt) +++ subList
 
 -- | Convert an Element to Html.
@@ -197,18 +238,32 @@ elementToHtml opts (Sec level num id' title' elements) = do
   innerContents <- mapM (elementToHtml opts) elements
   modify $ \st -> st{stSecNum = num}  -- update section number
   header' <- blockToHtml opts (Header level title')
-  return $ if writerS5 opts || (writerStrictMarkdown opts && not (writerTableOfContents opts))
-              -- S5 gets confused by the extra divs around sections
-              then toHtmlFromList (header' : innerContents)
-              else thediv ! [prefixedId opts id'] << (header' : innerContents)
+  let slides = writerSlideVariant opts `elem` [SlidySlides, S5Slides]
+  let header'' = header' !  [prefixedId opts id' |
+                             not (writerStrictMarkdown opts ||
+                                  writerSectionDivs opts || slides)]
+  let stuff = header'' : innerContents
+  return $ if slides   -- S5 gets confused by the extra divs around sections
+              then toHtmlFromList $ intersperse (nl opts) stuff
+              else if writerSectionDivs opts
+                      then if writerHtml5 opts
+                              then tag "section" ! [prefixedId opts id']
+                                     << (nl opts : (intersperse (nl opts) stuff
+                                                    ++ [nl opts]))
+                              else thediv ! [prefixedId opts id'] <<
+                                      (nl opts : (intersperse (nl opts) stuff
+                                                  ++ [nl opts]))
+                      else toHtmlFromList $ intersperse (nl opts) stuff
 
 -- | Convert list of Note blocks to a footnote <div>.
 -- Assumes notes are sorted.
-footnoteSection :: [Html] -> Html
-footnoteSection notes =
+footnoteSection :: WriterOptions -> [Html] -> Html
+footnoteSection opts notes =
   if null notes 
      then noHtml
-     else thediv ! [theclass "footnotes"] $ hr +++ (olist << notes)
+     else nl opts +++ (thediv ! [theclass "footnotes"]
+          $ nl opts +++ hr +++ nl opts +++
+            (olist << (notes ++ [nl opts])) +++ nl opts)
 
 
 -- | Parse a mailto link; return Just (name, domain) or Nothing.
@@ -249,7 +304,7 @@ obfuscateLink opts txt s =
                      linkText  ++ "+'<\\/'+'a'+'>');\n// -->\n")) +++  
                      noscript (primHtml $ obfuscateString altText)
                 _ -> error $ "Unknown obfuscation method: " ++ show meth
-        _ -> anchor ! [href s] $ stringToHtml txt  -- malformed email
+        _ -> anchor ! [href s] $ stringToHtml opts txt  -- malformed email
 
 -- | Obfuscate character as entity.
 obfuscateChar :: Char -> String
@@ -262,38 +317,53 @@ obfuscateChar char =
 obfuscateString :: String -> String
 obfuscateString = concatMap obfuscateChar . decodeCharacterReferences
 
+attrsToHtml :: WriterOptions -> Attr -> [HtmlAttr]
+attrsToHtml opts (id',classes',keyvals) =
+  [theclass (unwords classes') | not (null classes')] ++
+  [prefixedId opts id' | not (null id')] ++
+  map (\(x,y) -> strAttr x y) keyvals
+
 -- | Convert Pandoc block element to HTML.
 blockToHtml :: WriterOptions -> Block -> State WriterState Html
-blockToHtml _ Null = return $ noHtml 
+blockToHtml _ Null = return noHtml
 blockToHtml opts (Plain lst) = inlineListToHtml opts lst
 blockToHtml opts (Para [Image txt (s,tit)]) = do
   img <- inlineToHtml opts (Image txt (s,tit))
   capt <- inlineListToHtml opts txt
-  return $ thediv ! [theclass "figure"] <<
-             [img, paragraph ! [theclass "caption"] << capt]
-blockToHtml opts (Para lst) = inlineListToHtml opts lst >>= (return . paragraph)
-blockToHtml _ (RawHtml str) = return $ primHtml str
-blockToHtml _ (HorizontalRule) = return $ hr
+  return $ if writerHtml5 opts
+              then tag "figure" <<
+                    [nl opts, img, tag "figcaption" << capt, nl opts]
+              else thediv ! [theclass "figure"] <<
+                    [nl opts, img, paragraph ! [theclass "caption"] << capt,
+                    nl opts]
+blockToHtml opts (Para lst) = do
+  contents <- inlineListToHtml opts lst
+  return $ paragraph contents
+blockToHtml _ (RawBlock "html" str) = return $ primHtml str
+blockToHtml _ (RawBlock _ _) = return noHtml
+blockToHtml _ (HorizontalRule) = return hr
 blockToHtml opts (CodeBlock (id',classes,keyvals) rawCode) = do
   let classes' = if writerLiterateHaskell opts
                     then classes
                     else filter (/= "literate") classes
-  case highlightHtml (id',classes',keyvals) rawCode of
+  case highlightHtml False (id',classes',keyvals) rawCode of
          Left _  -> -- change leading newlines into <br /> tags, because some
                     -- browsers ignore leading newlines in pre blocks
                     let (leadingBreaks, rawCode') = span (=='\n') rawCode
-                        attrs = [theclass (unwords classes') | not (null classes')] ++
-                                [prefixedId opts id' | not (null id')] ++
-                                map (\(x,y) -> strAttr x y) keyvals
+                        attrs = attrsToHtml opts (id', classes', keyvals)
+                        addBird = if "literate" `elem` classes'
+                                     then unlines . map ("> " ++) . lines
+                                     else unlines . lines
                     in  return $ pre ! attrs $ thecode <<
                                  (replicate (length leadingBreaks) br +++
-                                 [stringToHtml $ rawCode' ++ "\n"])
-         Right h -> modify (\st -> st{ stHighlighting = True }) >> return h
+                                 [stringToHtml opts $ addBird rawCode'])
+         Right h -> modify (\st -> st{ stHighlighting = True }) >>
+                    return h
 blockToHtml opts (BlockQuote blocks) =
   -- in S5, treat list in blockquote specially
   -- if default is incremental, make it nonincremental; 
   -- otherwise incremental
-  if writerS5 opts
+  if writerSlideVariant opts /= NoSlides
      then let inc = not (writerIncremental opts) in
           case blocks of 
              [BulletList lst]  -> blockToHtml (opts {writerIncremental = inc})
@@ -301,33 +371,36 @@ blockToHtml opts (BlockQuote blocks) =
              [OrderedList attribs lst] -> 
                                   blockToHtml (opts {writerIncremental = inc})
                                   (OrderedList attribs lst)
-             _                 -> blockListToHtml opts blocks >>= 
-                                  (return . blockquote)
-     else blockListToHtml opts blocks >>= (return . blockquote)
+             _                 -> do contents <- blockListToHtml opts blocks
+                                     return $ blockquote (nl opts +++
+                                               contents +++ nl opts)
+     else do
+       contents <- blockListToHtml opts blocks
+       return $ blockquote (nl opts +++ contents +++ nl opts)
 blockToHtml opts (Header level lst) = do 
   contents <- inlineListToHtml opts lst
   secnum <- liftM stSecNum get
   let contents' = if writerNumberSections opts
                      then (thespan ! [theclass "header-section-number"] << showSecNum secnum) +++
-                            stringToHtml " " +++ contents
+                            stringToHtml opts " " +++ contents
                      else contents
   let contents''  = if writerTableOfContents opts
                        then anchor ! [href $ "#" ++ writerIdentifierPrefix opts ++ "TOC"] $ contents'
                        else contents'
-  return $ case level of
+  return $ (case level of
               1 -> h1 contents''
               2 -> h2 contents''
               3 -> h3 contents''
               4 -> h4 contents''
               5 -> h5 contents''
               6 -> h6 contents''
-              _ -> paragraph contents''
+              _ -> paragraph contents'')
 blockToHtml opts (BulletList lst) = do
   contents <- mapM (blockListToHtml opts) lst
   let attribs = if writerIncremental opts
                    then [theclass "incremental"]
                    else []
-  return $ unordList ! attribs $ contents
+  return $ (unordList opts contents) ! attribs
 blockToHtml opts (OrderedList (startnum, numstyle, _) lst) = do
   contents <- mapM (blockListToHtml opts) lst
   let numstyle' = camelCaseToHyphenated $ show numstyle
@@ -338,41 +411,60 @@ blockToHtml opts (OrderedList (startnum, numstyle, _) lst) = do
                    then [start startnum]
                    else []) ++
                 (if numstyle /= DefaultStyle
-                   then [thestyle $ "list-style-type: " ++ numstyle' ++ ";"]
+                   then if writerHtml5 opts
+                           then [strAttr "type" $
+                                 case numstyle of
+                                      Decimal    -> "1"
+                                      LowerAlpha -> "a"
+                                      UpperAlpha -> "A"
+                                      LowerRoman -> "i"
+                                      UpperRoman -> "I"
+                                      _          -> "1"]
+                           else [thestyle $ "list-style-type: " ++
+                                   numstyle']
                    else [])
-  return $ ordList ! attribs $ contents
+  return $ (ordList opts contents) ! attribs
 blockToHtml opts (DefinitionList lst) = do
   contents <- mapM (\(term, defs) ->
                   do term' <- liftM (dterm <<) $ inlineListToHtml opts term
-                     defs' <- mapM (liftM (ddef <<) . blockListToHtml opts) defs
-                     return $ term' : defs') lst
+                     defs' <- mapM ((liftM (\x -> ddef << (x +++ nl opts))) .
+                                    blockListToHtml opts) defs
+                     return $ nl opts : term' : nl opts : defs') lst
   let attribs = if writerIncremental opts
                    then [theclass "incremental"]
                    else []
-  return $ dlist ! attribs << concat contents
+  return $ dlist ! attribs << (concat contents +++ nl opts)
 blockToHtml opts (Table capt aligns widths headers rows') = do
-  let alignStrings = map alignmentToString aligns
   captionDoc <- if null capt
                    then return noHtml
-                   else inlineListToHtml opts capt >>= return . caption
+                   else do
+                     cs <- inlineListToHtml opts capt
+                     return $ caption cs +++ nl opts
   let percent w = show (truncate (100*w) :: Integer) ++ "%"
+  let widthAttrs w = if writerHtml5 opts
+                        then [thestyle $ "width: " ++ percent w]
+                        else [width $ percent w]
   let coltags = if all (== 0.0) widths
                    then noHtml
                    else concatHtml $ map
-                         (\w -> col ! [width $ percent w] $ noHtml) widths
+                         (\w -> (col ! (widthAttrs w)) noHtml +++ nl opts)
+                         widths
   head' <- if all null headers
               then return noHtml
-              else liftM (thead <<) $ tableRowToHtml opts alignStrings 0 headers
-  body' <- liftM (tbody <<) $
-               zipWithM (tableRowToHtml opts alignStrings) [1..] rows'
-  return $ table $ captionDoc +++ coltags +++ head' +++ body'
+              else do
+                contents <- tableRowToHtml opts aligns 0 headers
+                return $ thead << (nl opts +++ contents) +++ nl opts
+  body' <- liftM (\x -> tbody << (nl opts +++ x)) $
+               zipWithM (tableRowToHtml opts aligns) [1..] rows'
+  return $ table $ nl opts +++ captionDoc +++ coltags +++ head' +++
+                   body' +++ nl opts
 
 tableRowToHtml :: WriterOptions
-               -> [String]
+               -> [Alignment]
                -> Int
                -> [[Block]]
                -> State WriterState Html
-tableRowToHtml opts alignStrings rownum cols' = do
+tableRowToHtml opts aligns rownum cols' = do
   let mkcell = if rownum == 0 then th else td
   let rowclass = case rownum of
                       0                  -> "header"
@@ -380,8 +472,9 @@ tableRowToHtml opts alignStrings rownum cols' = do
                       _                  -> "even"
   cols'' <- sequence $ zipWith 
             (\alignment item -> tableItemToHtml opts mkcell alignment item) 
-            alignStrings cols'
-  return $ tr ! [theclass rowclass] $ toHtmlFromList cols''
+            aligns cols'
+  return $ (tr ! [theclass rowclass] $ nl opts +++ toHtmlFromList cols'')
+          +++ nl opts
 
 alignmentToString :: Alignment -> [Char]
 alignmentToString alignment = case alignment of
@@ -392,16 +485,26 @@ alignmentToString alignment = case alignment of
 
 tableItemToHtml :: WriterOptions
                 -> (Html -> Html)
-                -> [Char]
+                -> Alignment
                 -> [Block]
                 -> State WriterState Html
 tableItemToHtml opts tag' align' item = do
   contents <- blockListToHtml opts item
-  return $ tag' ! [align align'] $ contents
+  let alignAttrs = if writerHtml5 opts
+                      then [thestyle $ "align: " ++ alignmentToString align']
+                      else [align $ alignmentToString align']
+  return $ (tag' ! alignAttrs) contents +++ nl opts
+
+toListItems :: WriterOptions -> [Html] -> [Html]
+toListItems opts items = map (toListItem opts) items ++ [nl opts]
+
+toListItem :: WriterOptions -> Html -> Html
+toListItem opts item = nl opts +++ li item
 
 blockListToHtml :: WriterOptions -> [Block] -> State WriterState Html
-blockListToHtml opts lst = 
-  mapM (blockToHtml opts) lst >>= return . toHtmlFromList
+blockListToHtml opts lst =
+  mapM (blockToHtml opts) lst >>=
+  return . toHtmlFromList . intersperse (nl opts)
 
 -- | Convert list of Pandoc inline elements to HTML.
 inlineListToHtml :: WriterOptions -> [Inline] -> State WriterState Html
@@ -412,16 +515,20 @@ inlineListToHtml opts lst =
 inlineToHtml :: WriterOptions -> Inline -> State WriterState Html
 inlineToHtml opts inline =
   case inline of  
-    (Str str)        -> return $ stringToHtml str
-    (Space)          -> return $ stringToHtml " "
+    (Str str)        -> return $ stringToHtml opts str
+    (Space)          -> return $ stringToHtml opts " "
     (LineBreak)      -> return br
-    (EmDash)         -> return $ stringToHtml "—"
-    (EnDash)         -> return $ stringToHtml "–"
-    (Ellipses)       -> return $ stringToHtml "…"
-    (Apostrophe)     -> return $ stringToHtml "’"
+    (EmDash)         -> return $ stringToHtml opts "—"
+    (EnDash)         -> return $ stringToHtml opts "–"
+    (Ellipses)       -> return $ stringToHtml opts "…"
+    (Apostrophe)     -> return $ stringToHtml opts "’"
     (Emph lst)       -> inlineListToHtml opts lst >>= return . emphasize
     (Strong lst)     -> inlineListToHtml opts lst >>= return . strong
-    (Code str)       -> return $ thecode << str
+    (Code attr str)  -> case highlightHtml True attr str of
+                             Left _  -> return
+                                        $ thecode ! (attrsToHtml opts attr)
+                                        $ stringToHtml opts str
+                             Right h -> return h
     (Strikeout lst)  -> inlineListToHtml opts lst >>=
                         return . (thespan ! [thestyle "text-decoration: line-through;"])
     (SmallCaps lst)   -> inlineListToHtml opts lst >>=
@@ -430,32 +537,37 @@ inlineToHtml opts inline =
     (Subscript lst)   -> inlineListToHtml opts lst >>= return . sub
     (Quoted quoteType lst) ->
                         let (leftQuote, rightQuote) = case quoteType of
-                              SingleQuote -> (stringToHtml "‘",
-                                              stringToHtml "’")
-                              DoubleQuote -> (stringToHtml "“",
-                                              stringToHtml "”")
+                              SingleQuote -> (stringToHtml opts "‘",
+                                              stringToHtml opts "’")
+                              DoubleQuote -> (stringToHtml opts "“",
+                                              stringToHtml opts "”")
                         in  do contents <- inlineListToHtml opts lst
                                return $ leftQuote +++ contents +++ rightQuote
-    (Math t str) -> 
-                        modify (\st -> st {stMath = True}) >> 
+    (Math t str) ->     modify (\st -> st {stMath = True}) >> 
                         (case writerHTMLMathMethod opts of
                                LaTeXMathML _ -> 
                                   -- putting LaTeXMathML in container with class "LaTeX" prevents
                                   -- non-math elements on the page from being treated as math by
                                   -- the javascript
                                   return $ thespan ! [theclass "LaTeX"] $
-                                             if t == InlineMath
-                                                 then primHtml ("$" ++ str ++ "$")
-                                                 else primHtml ("$$" ++ str ++ "$$")
-                               JsMath _ ->
-                                  return $ if t == InlineMath
-                                              then thespan ! [theclass "math"] $ primHtml str
-                                              else thediv ! [theclass "math"]  $ primHtml str
-                               MimeTeX url -> 
-                                  return $ image ! [src (url ++ "?" ++ str),
-                                                    alt str, title str]
+                                         case t of
+                                           InlineMath  -> primHtml ("$" ++ str ++ "$")
+                                           DisplayMath -> primHtml ("$$" ++ str ++ "$$")
+                               JsMath _ -> do
+                                  let m = primHtml str
+                                  return $ case t of
+                                           InlineMath -> thespan ! [theclass "math"] $ m
+                                           DisplayMath -> thediv ! [theclass "math"] $ m
+                               WebTeX url -> do
+                                  let m = image ! [src (url ++ urlEncode str),
+                                                         alt str, title str]
+                                  return $ case t of
+                                            InlineMath  -> m
+                                            DisplayMath -> br +++ m +++ br
                                GladTeX ->
-                                  return $ primHtml $ "<EQ>" ++ str ++ "</EQ>"
+                                  return $ case t of
+                                             InlineMath -> primHtml $ "<EQ ENV=\"math\">" ++ str ++ "</EQ>"
+                                             DisplayMath -> primHtml $ "<EQ ENV=\"displaymath\">" ++ str ++ "</EQ>"
                                MathML _ -> do
                                   let dt = if t == InlineMath
                                               then DisplayInline
@@ -466,18 +578,25 @@ inlineToHtml opts inline =
                                         Right r -> return $ primHtml $
                                                     ppcElement conf r
                                         Left  _ -> inlineListToHtml opts
-                                                    (readTeXMath str) >>=
-                                                      return . (thespan !
-                                                      [theclass "math"])
-                               PlainMath ->
-                                  inlineListToHtml opts (readTeXMath str) >>=
-                                  return . (thespan ! [theclass "math"]) ) 
-    (TeX str)        -> case writerHTMLMathMethod opts of
-                              LaTeXMathML _ -> do modify (\st -> st {stMath = True})
-                                                  return $ primHtml str
-                              _             -> return noHtml
-    (HtmlInline str) -> return $ primHtml str 
-    (Link [Code str] (s,_)) | "mailto:" `isPrefixOf` s ->
+                                                   (readTeXMath str) >>= return .
+                                                     (thespan !  [theclass "math"])
+                               MathJax _ -> return $ primHtml $
+                                  case t of
+                                    InlineMath  -> "\\(" ++ str ++ "\\)"
+                                    DisplayMath -> "\\[" ++ str ++ "\\]"
+                               PlainMath -> do
+                                  x <- inlineListToHtml opts (readTeXMath str)
+                                  let m = thespan ! [theclass "math"] $ x
+                                  return  $ case t of
+                                             InlineMath  -> m
+                                             DisplayMath -> br +++ m +++ br )
+    (RawInline "latex" str) -> case writerHTMLMathMethod opts of
+                               LaTeXMathML _ -> do modify (\st -> st {stMath = True})
+                                                   return $ primHtml str
+                               _             -> return noHtml
+    (RawInline "html" str) -> return $ primHtml str 
+    (RawInline _ _) -> return noHtml
+    (Link [Code _ str] (s,_)) | "mailto:" `isPrefixOf` s ->
                         return $ obfuscateLink opts str s
     (Link txt (s,_)) | "mailto:" `isPrefixOf` s -> do
                         linkText <- inlineListToHtml opts txt  
@@ -488,8 +607,7 @@ inlineToHtml opts inline =
                                  if null tit then [] else [title tit]) $ 
                                  linkText
     (Image txt (s,tit)) -> do
-                        alternate <- inlineListToHtml opts txt
-                        let alternate' = renderFragment opts alternate
+                        let alternate' = stringify txt
                         let attributes = [src s] ++
                                          (if null tit 
                                             then [] 
@@ -517,9 +635,10 @@ blockListToNote :: WriterOptions -> String -> [Block] -> State WriterState Html
 blockListToNote opts ref blocks =
   -- If last block is Para or Plain, include the backlink at the end of
   -- that block. Otherwise, insert a new Plain block with the backlink.
-  let backlink = [HtmlInline $ " <a href=\"#" ++ writerIdentifierPrefix opts ++ "fnref" ++ ref ++ 
+  let backlink = [RawInline "html" $ " <a href=\"#" ++ writerIdentifierPrefix opts ++ "fnref" ++ ref ++ 
                  "\" class=\"footnoteBackLink\"" ++
-                 " title=\"Jump back to footnote " ++ ref ++ "\">↩</a>"]
+                 " title=\"Jump back to footnote " ++ ref ++ "\">" ++
+                 (if writerAscii opts then "&#8617;" else "↩") ++ "</a>"]
       blocks'  = if null blocks
                     then []
                     else let lastBlock   = last blocks
@@ -532,5 +651,5 @@ blockListToNote opts ref blocks =
                                   _           -> otherBlocks ++ [lastBlock,
                                                  Plain backlink]
   in  do contents <- blockListToHtml opts blocks'
-         return $ li ! [prefixedId opts ("fn" ++ ref)] $ contents
+         return $ nl opts +++ (li ! [prefixedId opts ("fn" ++ ref)]) contents
 
