@@ -35,7 +35,7 @@ import Text.Pandoc.Shared
 import Text.Pandoc.XML
 import Text.Pandoc.Templates (renderTemplate)
 import Text.Pandoc.Readers.TeXMath
-import Text.PrettyPrint.HughesPJ hiding ( Str )
+import Text.Pandoc.Pretty
 import Text.Printf ( printf )
 import Control.Applicative ( (<$>) )
 import Control.Arrow ( (***), (>>>) )
@@ -112,7 +112,9 @@ setInDefinitionList :: Bool -> State WriterState ()
 setInDefinitionList b = modify $  \s -> s { stInDefinition = b }
 
 inParagraphTags :: Doc -> Doc
-inParagraphTags = inTags False "text:p" [("text:style-name", "Text_20_body")]
+inParagraphTags d | isEmpty d = empty
+inParagraphTags d =
+  inTags False "text:p" [("text:style-name", "Text_20_body")] d
 
 inParagraphTagsWithStyle :: String -> Doc -> Doc
 inParagraphTagsWithStyle sty = inTags False "text:p" [("text:style-name", sty)]
@@ -167,7 +169,11 @@ writeOpenDocument opts (Pandoc (Meta title authors date) blocks) =
            date'' <- inlinesToOpenDocument opts date
            doc'' <- blocksToOpenDocument opts blocks
            return (doc'', title'', authors'', date'')
-      body'    = render doc
+      colwidth = if writerWrapText opts
+                    then Just $ writerColumns opts
+                    else Nothing
+      render' = render colwidth
+      body'    = render' doc
       styles   = stTableStyles s ++ stParaStyles s ++ stTextStyles s
       listStyle (n,l) = inTags True "text:list-style"
                           [("style:name", "L" ++ show n)] (vcat l)
@@ -176,10 +182,10 @@ writeOpenDocument opts (Pandoc (Meta title authors date) blocks) =
                           reverse $ styles ++ listStyles
       context = writerVariables opts ++
                 [ ("body", body')
-                , ("automatic-styles", render automaticStyles)
-                , ("title", render title')
-                , ("date", render date') ] ++
-                [ ("author", render a) | a <- authors' ]
+                , ("automatic-styles", render' automaticStyles)
+                , ("title", render' title')
+                , ("date", render' date') ] ++
+                [ ("author", render' a) | a <- authors' ]
   in  if writerStandalone opts
          then renderTemplate context $ writerTemplate opts
          else body'
@@ -273,7 +279,7 @@ blockToOpenDocument o bs
     | Header       i b <- bs = inHeaderTags  i <$> inlinesToOpenDocument o b
     | BlockQuote     b <- bs = mkBlockQuote b
     | CodeBlock    _ s <- bs = preformatted s
-    | RawHtml        _ <- bs = return empty
+    | RawBlock _     _ <- bs = return empty
     | DefinitionList b <- bs = defList b
     | BulletList     b <- bs = bulletListToOpenDocument o b
     | OrderedList  a b <- bs = orderedList a b
@@ -286,7 +292,7 @@ blockToOpenDocument o bs
                            r <- vcat  <$> mapM (deflistItemToOpenDocument o) b
                            setInDefinitionList False
                            return r
-      preformatted  s = vcat <$> mapM (inPreformattedTags . escapeStringForXML) (lines s)
+      preformatted  s = (flush . vcat) <$> mapM (inPreformattedTags . escapeStringForXML) (lines s)
       mkBlockQuote  b = do increaseIndent
                            i <- paraStyle "Quotations" []
                            inBlockQuote o i (map plainToPara b)
@@ -346,7 +352,7 @@ inlineToOpenDocument o ils
     | EmDash        <- ils = inTextStyle $ text "&#8212;"
     | EnDash        <- ils = inTextStyle $ text "&#8211;"
     | Apostrophe    <- ils = inTextStyle $ text "&#8217;"
-    | Space         <- ils = inTextStyle $ char ' '
+    | Space         <- ils = inTextStyle space
     | LineBreak     <- ils = return $ selfClosingTag "text:line-break" []
     | Str         s <- ils = inTextStyle $ handleSpaces $ escapeStringForXML s
     | Emph        l <- ils = withTextStyle Italic $ inlinesToOpenDocument o l
@@ -356,11 +362,12 @@ inlineToOpenDocument o ils
     | Subscript   l <- ils = withTextStyle Sub    $ inlinesToOpenDocument o l
     | SmallCaps   l <- ils = withTextStyle SmallC $ inlinesToOpenDocument o l
     | Quoted    t l <- ils = inQuotes t <$> inlinesToOpenDocument o l
-    | Code        s <- ils = preformatted s
+    | Code      _ s <- ils = preformatted s
     | Math      _ s <- ils = inlinesToOpenDocument o (readTeXMath s)
     | Cite      _ l <- ils = inlinesToOpenDocument o l
-    | TeX         s <- ils = preformatted s
-    | HtmlInline  s <- ils = preformatted s
+    | RawInline "opendocument" s <- ils = preformatted s
+    | RawInline "html" s <- ils = preformatted s  -- for backwards compat.
+    | RawInline _ _ <- ils = return empty
     | Link  l (s,t) <- ils = mkLink s t <$> inlinesToOpenDocument o l
     | Image _ (s,_) <- ils = return $ mkImg  s
     | Note        l <- ils = mkNote l
