@@ -34,7 +34,8 @@ module Text.Pandoc.XML ( stripTags,
                          selfClosingTag,
                          inTagsSimple,
                          inTagsIndented  ) where
-import Text.PrettyPrint.HughesPJ
+
+import Text.Pandoc.Pretty
 
 -- | Remove everything between <...>
 stripTags :: String -> String
@@ -55,23 +56,15 @@ escapeCharForXML x = case x of
                        '"'    -> "&quot;"
                        c      -> [c]
 
--- | True if the character needs to be escaped.
-needsEscaping :: Char -> Bool
-needsEscaping c = c `elem` "&<>\""
-
 -- | Escape string as needed for XML.  Entity references are not preserved.
 escapeStringForXML :: String -> String
-escapeStringForXML ""  = ""
-escapeStringForXML str =
-  case break needsEscaping str of
-    (okay, "")     -> okay
-    (okay, (c:cs)) -> okay ++ escapeCharForXML c ++ escapeStringForXML cs
+escapeStringForXML = concatMap escapeCharForXML
 
 -- | Return a text object with a string of formatted XML attributes.
 attributeList :: [(String, String)] -> Doc
-attributeList = text .  concatMap
-  (\(a, b) -> " " ++ escapeStringForXML a ++ "=\"" ++
-  escapeStringForXML b ++ "\"")
+attributeList = hcat . map
+  (\(a, b) -> text (' ' : escapeStringForXML a ++ "=\"" ++
+  escapeStringForXML b ++ "\""))
 
 -- | Put the supplied contents between start and end tags of tagType,
 --   with specified attributes and (if specified) indentation.
