@@ -1,5 +1,5 @@
 {-
-Copyright (C) 2006-2010 John MacFarlane <jgm@berkeley.edu>
+Copyright (C) 2006-2011 John MacFarlane <jgm@berkeley.edu>
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -18,7 +18,7 @@ Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
 {- |
    Module      : Main
-   Copyright   : Copyright (C) 2006-2010 John MacFarlane
+   Copyright   : Copyright (C) 2006-2011 John MacFarlane
    License     : GNU GPL, version 2 or above
 
    Maintainer  : John MacFarlane <jgm@berkeley@edu>
@@ -30,49 +30,39 @@ writers.
 -}
 module Main where
 import Text.Pandoc
-import Text.Pandoc.ODT
-import Text.Pandoc.Writers.S5 (s5HeaderIncludes)
-import Text.Pandoc.Shared ( tabFilter, ObfuscationMethod (..), readDataFile )
+import Text.Pandoc.S5 (s5HeaderIncludes)
+import Text.Pandoc.Shared ( tabFilter, ObfuscationMethod (..), readDataFile,
+                            headerShift, findDataFile, normalize )
 #ifdef _HIGHLIGHTING
 import Text.Pandoc.Highlighting ( languages )
 #endif
-import System.Environment ( getArgs, getProgName, getEnvironment )
+import System.Environment ( getArgs, getProgName )
 import System.Exit ( exitWith, ExitCode (..) )
 import System.FilePath
 import System.Console.GetOpt
-import Data.Maybe ( fromMaybe )
-import Data.Char ( toLower, isDigit )
-import Data.List ( intercalate, isSuffixOf )
-import System.Directory ( getAppUserDataDirectory )
+import Data.Char ( toLower )
+import Data.List ( intercalate, isSuffixOf, isPrefixOf )
+import System.Directory ( getAppUserDataDirectory, doesFileExist )
 import System.IO ( stdout, stderr )
--- Note: ghc >= 6.12 (base >=4.2) supports unicode through iconv
--- So we use System.IO.UTF8 only if we have an earlier version
-#if MIN_VERSION_base(4,2,0)
-import System.IO ( hPutStr, hPutStrLn )
-#else
-import Prelude hiding ( putStr, putStrLn, writeFile, readFile, getContents )
-import System.IO.UTF8
-#endif
-#ifdef _CITEPROC
+import qualified Text.Pandoc.UTF8 as UTF8
 import Text.CSL
 import Text.Pandoc.Biblio
-#endif
 import Control.Monad (when, unless, liftM)
 import Network.HTTP (simpleHTTP, mkRequest, getResponseBody, RequestMethod(..))
-import Network.URI (parseURI, isURI)
-import Data.ByteString.Lazy.UTF8 (toString)
+import Network.URI (parseURI, isURI, URI(..))
+import qualified Data.ByteString.Lazy as B
+import Data.ByteString.Lazy.UTF8 (toString )
+import Codec.Binary.UTF8.String (decodeString, encodeString)
 
 copyrightMessage :: String
-copyrightMessage = "\nCopyright (C) 2006-2010 John MacFarlane\n" ++
+copyrightMessage = "\nCopyright (C) 2006-2011 John MacFarlane\n" ++
                     "Web:  http://johnmacfarlane.net/pandoc\n" ++
                     "This is free software; see the source for copying conditions.  There is no\n" ++
                     "warranty, not even for merchantability or fitness for a particular purpose."
 
 compileInfo :: String
 compileInfo =
-#ifdef _CITEPROC
   "\nCompiled with citeproc support." ++
-#endif
 #ifdef _HIGHLIGHTING
    "\nCompiled with syntax highlighting support for:\n" ++
        wrapWords 78 languages ++
@@ -90,57 +80,8 @@ wrapWords c = wrap' c c where
                                                        then ",\n" ++ x ++ wrap' cols (cols - length x) xs
                                                        else ", "  ++ x ++ wrap' cols (remaining - (length x + 2)) xs
 
--- | Association list of formats and readers.
-readers :: [(String, ParserState -> String -> Pandoc)]
-readers = [("native"       , readPandoc)
-          ,("markdown"     , readMarkdown)
-          ,("markdown+lhs" , readMarkdown)
-          ,("rst"          , readRST)
-          ,("rst+lhs"      , readRST)
-          ,("html"         , readHtml)
-          ,("latex"        , readLaTeX)
-          ,("latex+lhs"    , readLaTeX)
-          ]
-
--- | Reader for native Pandoc format.
-readPandoc :: ParserState -> String -> Pandoc
-readPandoc _ = read
-
--- | Association list of formats and writers.
-writers :: [ ( String, WriterOptions -> Pandoc -> String ) ]
-writers = [("native"       , writeDoc)
-          ,("html"         , writeHtmlString)
-          ,("html+lhs"     , writeHtmlString)
-          ,("s5"           , writeS5String)
-          ,("docbook"      , writeDocbook)
-          ,("opendocument" , writeOpenDocument)
-          ,("odt"          , writeOpenDocument)
-          ,("latex"        , writeLaTeX)
-          ,("latex+lhs"    , writeLaTeX)
-          ,("context"      , writeConTeXt)
-          ,("texinfo"      , writeTexinfo)
-          ,("man"          , writeMan)
-          ,("markdown"     , writeMarkdown)
-          ,("markdown+lhs" , writeMarkdown)
-          ,("plain"        , writePlain)
-          ,("rst"          , writeRST)
-          ,("rst+lhs"      , writeRST)
-          ,("mediawiki"    , writeMediaWiki)
-          ,("rtf"          , writeRTF)
-          ]
-
 isNonTextOutput :: String -> Bool
-isNonTextOutput = (`elem` ["odt"])
-
--- | Writer for Pandoc native format.
-writeDoc :: WriterOptions -> Pandoc -> String
-writeDoc _ = prettyPandoc
-
-headerShift :: Int -> Pandoc -> Pandoc
-headerShift n = processWith shift
-  where shift :: Block -> Block
-        shift (Header level inner) = Header (level + n) inner
-        shift x                    = x
+isNonTextOutput = (`elem` ["odt","epub"])
 
 -- | Data structure for command line options.
 data Opt = Opt
@@ -154,31 +95,35 @@ data Opt = Opt
     , optTransforms        :: [Pandoc -> Pandoc]  -- ^ Doc transforms to apply
     , optTemplate          :: String  -- ^ Custom template
     , optVariables         :: [(String,String)] -- ^ Template variables to set
-    , optBefore            :: [String] -- ^ Texts to include before body
-    , optAfter             :: [String] -- ^ Texts to include after body
     , optOutputFile        :: String  -- ^ Name of output file
     , optNumberSections    :: Bool    -- ^ Number sections in LaTeX
-    , optIncremental       :: Bool    -- ^ Use incremental lists in S5
+    , optSectionDivs       :: Bool    -- ^ Put sections in div tags in HTML
+    , optIncremental       :: Bool    -- ^ Use incremental lists in Slidy/S5
+    , optOffline           :: Bool    -- ^ Make slideshow accessible offline
     , optXeTeX             :: Bool    -- ^ Format latex for xetex
     , optSmart             :: Bool    -- ^ Use smart typography
+    , optHtml5             :: Bool    -- ^ Produce HTML5 in HTML
+    , optChapters          :: Bool    -- ^ Use chapter for top-level sects
     , optHTMLMathMethod    :: HTMLMathMethod -- ^ Method to print HTML math
     , optReferenceODT      :: Maybe FilePath -- ^ Path of reference.odt
+    , optEPUBStylesheet    :: Maybe String   -- ^ EPUB stylesheet
+    , optEPUBMetadata      :: String  -- ^ EPUB metadata
     , optDumpArgs          :: Bool    -- ^ Output command-line arguments
     , optIgnoreArgs        :: Bool    -- ^ Ignore command-line arguments
     , optStrict            :: Bool    -- ^ Use strict markdown syntax
     , optReferenceLinks    :: Bool    -- ^ Use reference links in writing markdown, rst
     , optWrapText          :: Bool    -- ^ Wrap text
-    , optSanitizeHTML      :: Bool    -- ^ Sanitize HTML
+    , optColumns           :: Int     -- ^ Line length in characters
     , optPlugins           :: [Pandoc -> IO Pandoc] -- ^ Plugins to apply
     , optEmailObfuscation  :: ObfuscationMethod
     , optIdentifierPrefix  :: String
     , optIndentedCodeClasses :: [String] -- ^ Default classes for indented code blocks
     , optDataDir           :: Maybe FilePath
-#ifdef _CITEPROC
-    , optBiblioFile        :: String
-    , optBiblioFormat      :: String
-    , optCslFile           :: String
-#endif
+    , optCiteMethod        :: CiteMethod -- ^ Method to output cites
+    , optBibliography      :: [String]
+    , optCslFile           :: FilePath
+    , optListings          :: Bool       -- ^ Use listings package for code blocks
+    , optAscii             :: Bool       -- ^ Avoid using nonascii characters
     }
 
 -- | Defaults for command-line options.
@@ -194,31 +139,35 @@ defaultOpts = Opt
     , optTransforms        = []
     , optTemplate          = ""
     , optVariables         = []
-    , optBefore            = []
-    , optAfter             = []
     , optOutputFile        = "-"    -- "-" means stdout
     , optNumberSections    = False
+    , optSectionDivs       = False
     , optIncremental       = False
+    , optOffline           = False
     , optXeTeX             = False
     , optSmart             = False
+    , optHtml5             = False
+    , optChapters          = False
     , optHTMLMathMethod    = PlainMath
     , optReferenceODT      = Nothing
+    , optEPUBStylesheet    = Nothing
+    , optEPUBMetadata      = ""
     , optDumpArgs          = False
     , optIgnoreArgs        = False
     , optStrict            = False
     , optReferenceLinks    = False
     , optWrapText          = True
-    , optSanitizeHTML      = False
+    , optColumns           = 72
     , optPlugins           = []
     , optEmailObfuscation  = JavascriptObfuscation
     , optIdentifierPrefix  = ""
     , optIndentedCodeClasses = []
     , optDataDir           = Nothing
-#ifdef _CITEPROC
-    , optBiblioFile        = []
-    , optBiblioFormat      = []
-    , optCslFile           = []
-#endif
+    , optCiteMethod        = Citeproc
+    , optBibliography      = []
+    , optCslFile           = ""
+    , optListings          = False
+    , optAscii             = False
     }
 
 -- | A list of functions, each transforming the options data structure
@@ -229,13 +178,13 @@ options =
                  (ReqArg
                   (\arg opt -> return opt { optReader = map toLower arg })
                   "FORMAT")
-                 "" -- ("(" ++ (intercalate ", " $ map fst readers) ++ ")")
+                 ""
 
     , Option "tw" ["to","write"]
                  (ReqArg
                   (\arg opt -> return opt { optWriter = map toLower arg })
                   "FORMAT")
-                 "" -- ("(" ++ (intercalate ", " $ map fst writers) ++ ")")
+                 ""
 
     , Option "s" ["standalone"]
                  (NoArg
@@ -255,8 +204,14 @@ options =
 
     , Option "" ["tab-stop"]
                  (ReqArg
-                  (\arg opt -> return opt { optTabStop = (read arg) } )
-                  "TABSTOP")
+                  (\arg opt ->
+                      case reads arg of
+                           [(t,"")] | t > 0 -> return opt { optTabStop = t }
+                           _          -> do
+                               UTF8.hPutStrLn stderr $
+                                   "tab-stop must be a number greater than 0"
+                               exitWith $ ExitFailure 31)
+                  "NUMBER")
                  "" -- "Tab stop (default 4)"
 
     , Option "" ["strict"]
@@ -264,6 +219,12 @@ options =
                   (\opt -> return opt { optStrict = True } ))
                  "" -- "Disable markdown syntax extensions"
 
+    , Option "" ["normalize"]
+                 (NoArg
+                  (\opt -> return opt { optTransforms =
+                                   normalize : optTransforms opt } ))
+                 "" -- "Normalize the Pandoc AST"
+
     , Option "" ["reference-links"]
                  (NoArg
                   (\opt -> return opt { optReferenceLinks = True } ))
@@ -279,6 +240,11 @@ options =
                   (\opt -> return opt { optSmart = True }))
                  "" -- "Use smart quotes, dashes, and ellipses"
 
+    , Option "5" ["html5"]
+                 (NoArg
+                  (\opt -> return opt { optHtml5 = True }))
+                 "" -- "Produce HTML5 in HTML output"
+
     , Option "m" ["latexmathml", "asciimathml"]
                  (OptArg
                   (\arg opt ->
@@ -295,17 +261,36 @@ options =
 
     , Option "" ["mimetex"]
                  (OptArg
-                  (\arg opt -> return opt { optHTMLMathMethod = MimeTeX
-                                  (fromMaybe "/cgi-bin/mimetex.cgi" arg)})
+                  (\arg opt -> do
+                      let url' = case arg of
+                                      Just u   -> u ++ "?"
+                                      Nothing  -> "/cgi-bin/mimetex.cgi?"
+                      return opt { optHTMLMathMethod = WebTeX url' })
                   "URL")
                  "" -- "Use mimetex for HTML math"
 
+    , Option "" ["webtex"]
+                 (OptArg
+                  (\arg opt -> do
+                      let url' = case arg of
+                                      Just u   -> u
+                                      Nothing  -> "http://chart.apis.google.com/chart?cht=tx&chl="
+                      return opt { optHTMLMathMethod = WebTeX url' })
+                  "URL")
+                 "" -- "Use web service for HTML math"
+
     , Option "" ["jsmath"]
                  (OptArg
                   (\arg opt -> return opt { optHTMLMathMethod = JsMath arg})
                   "URL")
                  "" -- "Use jsMath for HTML math"
 
+    , Option "" ["mathjax"]
+                 (ReqArg
+                  (\arg opt -> return opt { optHTMLMathMethod = MathJax arg})
+                  "URL")
+                 "" -- "Use MathJax for HTML math"
+
     , Option "" ["gladtex"]
                  (NoArg
                   (\opt -> return opt { optHTMLMathMethod = GladTeX }))
@@ -314,27 +299,60 @@ options =
     , Option "i" ["incremental"]
                  (NoArg
                   (\opt -> return opt { optIncremental = True }))
-                 "" -- "Make list items display incrementally in S5"
+                 "" -- "Make list items display incrementally in Slidy/S5"
+
+    , Option "" ["offline"]
+                 (NoArg
+                  (\opt -> return opt { optOffline = True,
+                                        optStandalone = True }))
+                 "" -- "Make slide shows include all the needed js and css"
 
     , Option "" ["xetex"]
                  (NoArg
                   (\opt -> return opt { optXeTeX = True }))
                  "" -- "Format latex for processing by XeTeX"
 
+    , Option "" ["chapters"]
+                 (NoArg
+                  (\opt -> return opt { optChapters = True }))
+                 "" -- "Use chapter for top-level sections in LaTeX, DocBook"
+
     , Option "N" ["number-sections"]
                  (NoArg
                   (\opt -> return opt { optNumberSections = True }))
                  "" -- "Number sections in LaTeX"
 
+    , Option "" ["listings"]
+                 (NoArg
+                  (\opt -> return opt { optListings = True }))
+                 "" -- "Use listings package for LaTeX code blocks"
+
+    , Option "" ["section-divs"]
+                 (NoArg
+                  (\opt -> return opt { optSectionDivs = True }))
+                 "" -- "Put sections in div tags in HTML"
+
     , Option "" ["no-wrap"]
                  (NoArg
                   (\opt -> return opt { optWrapText = False }))
                  "" -- "Do not wrap text in output"
 
-    , Option "" ["sanitize-html"]
+    , Option "" ["columns"]
+                 (ReqArg
+                  (\arg opt ->
+                      case reads arg of
+                           [(t,"")] | t > 0 -> return opt { optColumns = t }
+                           _          -> do
+                               UTF8.hPutStrLn stderr $
+                                   "columns must be a number greater than 0"
+                               exitWith $ ExitFailure 33)
+                 "NUMBER")
+                 "" -- "Length of line in characters"
+
+    , Option "" ["ascii"]
                  (NoArg
-                  (\opt -> return opt { optSanitizeHTML = True }))
-                 "" -- "Sanitize HTML"
+                  (\opt -> return opt { optAscii = True }))
+                 "" -- "Avoid using non-ascii characters in output"
 
     , Option "" ["email-obfuscation"]
                  (ReqArg
@@ -343,7 +361,7 @@ options =
                             "references" -> return ReferenceObfuscation
                             "javascript" -> return JavascriptObfuscation
                             "none"       -> return NoObfuscation
-                            _            -> hPutStrLn stderr ("Error: Unknown obfuscation method: " ++ arg) >>
+                            _            -> UTF8.hPutStrLn stderr ("Error: Unknown obfuscation method: " ++ arg) >>
                                             exitWith (ExitFailure 6)
                      return opt { optEmailObfuscation = method })
                   "none|javascript|references")
@@ -369,23 +387,24 @@ options =
 
     , Option "" ["base-header-level"]
                  (ReqArg
-                  (\arg opt -> do
-                     if all isDigit arg && (read arg :: Int) >= 1
-                        then do
-                           let oldTransforms = optTransforms opt
-                           let shift = read arg - 1
-                           return opt{ optTransforms =
-                                         headerShift shift : oldTransforms }
-                        else do
-                           hPutStrLn stderr $ "base-header-level must be a number >= 1"
-                           exitWith $ ExitFailure 19)
-                  "LEVEL")
+                  (\arg opt ->
+                      case reads arg of
+                           [(t,"")] | t > 0 -> do
+                               let oldTransforms = optTransforms opt
+                               let shift = t - 1
+                               return opt{ optTransforms =
+                                           headerShift shift : oldTransforms }
+                           _          -> do
+                               UTF8.hPutStrLn stderr $
+                                   "base-header-level must be a number > 0"
+                               exitWith $ ExitFailure 19)
+                  "NUMBER")
                  "" -- "Headers base level"
 
     , Option "" ["template"]
                  (ReqArg
                   (\arg opt -> do
-                     text <- readFile arg
+                     text <- UTF8.readFile arg
                      return opt{ optTemplate = text,
                                  optStandalone = True })
                   "FILENAME")
@@ -399,9 +418,9 @@ options =
                             let newvars = optVariables opt ++ [(k,v)]
                             return opt{ optVariables = newvars }
                           _  -> do
-                            hPutStrLn stderr $ "Could not parse `" ++ arg ++ "' as a key/value pair (k=v or k:v)"
+                            UTF8.hPutStrLn stderr $ "Could not parse `" ++ arg ++ "' as a key/value pair (k=v or k:v)"
                             exitWith $ ExitFailure 17)
-                  "FILENAME")
+                  "KEY:VALUE")
                  "" -- "Use custom template"
 
     , Option "c" ["css"]
@@ -417,7 +436,7 @@ options =
     , Option "H" ["include-in-header"]
                  (ReqArg
                   (\arg opt -> do
-                     text <- readFile arg
+                     text <- UTF8.readFile arg
                      -- add new ones to end, so they're included in order specified
                      let newvars = optVariables opt ++ [("header-includes",text)]
                      return opt { optVariables = newvars,
@@ -428,7 +447,7 @@ options =
     , Option "B" ["include-before-body"]
                  (ReqArg
                   (\arg opt -> do
-                     text <- readFile arg
+                     text <- UTF8.readFile arg
                      -- add new ones to end, so they're included in order specified
                      let newvars = optVariables opt ++ [("include-before",text)]
                      return opt { optVariables = newvars,
@@ -439,7 +458,7 @@ options =
     , Option "A" ["include-after-body"]
                  (ReqArg
                   (\arg opt -> do
-                     text <- readFile arg
+                     text <- UTF8.readFile arg
                      -- add new ones to end, so they're included in order specified
                      let newvars = optVariables opt ++ [("include-after",text)]
                      return opt { optVariables = newvars,
@@ -447,16 +466,6 @@ options =
                   "FILENAME")
                  "" -- "File to include after document body"
 
-    , Option "C" ["custom-header"]
-                 (ReqArg
-                  (\arg opt -> do
-                     text <- readFile arg
-                     let newVars = ("legacy-header", text) : optVariables opt
-                     return opt { optVariables = newVars
-                                , optStandalone = True })
-                  "FILENAME")
-                 "" -- "File to use for custom header (implies -s)"
-
     , Option "T" ["title-prefix"]
                  (ReqArg
                   (\arg opt -> do
@@ -473,33 +482,55 @@ options =
                   "FILENAME")
                  "" -- "Path of custom reference.odt"
 
+    , Option "" ["epub-stylesheet"]
+                 (ReqArg
+                  (\arg opt -> do
+                     text <- UTF8.readFile arg
+                     return opt { optEPUBStylesheet = Just text })
+                  "FILENAME")
+                 "" -- "Path of epub.css"
+
+    , Option "" ["epub-metadata"]
+                 (ReqArg
+                  (\arg opt -> do
+                     text <- UTF8.readFile arg
+                     return opt { optEPUBMetadata = text })
+                  "FILENAME")
+                 "" -- "Path of epub metadata file"
+
     , Option "D" ["print-default-template"]
                  (ReqArg
                   (\arg _ -> do
                      templ <- getDefaultTemplate Nothing arg
                      case templ of
-                          Right t -> hPutStr stdout t
+                          Right t -> UTF8.hPutStr stdout t
                           Left e  -> error $ show e
                      exitWith ExitSuccess)
                   "FORMAT")
                  "" -- "Print default template for FORMAT"
-#ifdef _CITEPROC
-    , Option "" ["biblio"]
+
+    , Option "" ["bibliography"]
                  (ReqArg
-                  (\arg opt -> return opt { optBiblioFile = arg} )
+                  (\arg opt -> return opt { optBibliography = (optBibliography opt) ++ [arg] })
                   "FILENAME")
                  ""
-    , Option "" ["biblio-format"]
-                 (ReqArg
-                  (\arg opt -> return opt { optBiblioFormat = arg} )
-                  "STRING")
-                 ""
+
     , Option "" ["csl"]
                  (ReqArg
-                  (\arg opt -> return opt { optCslFile = arg} )
+                  (\arg opt -> return opt { optCslFile = arg })
                   "FILENAME")
                  ""
-#endif
+
+    , Option "" ["natbib"]
+                 (NoArg
+                  (\opt -> return opt { optCiteMethod = Natbib }))
+                 "" -- "Use natbib cite commands in LaTeX output"
+
+    , Option "" ["biblatex"]
+                 (NoArg
+                  (\opt -> return opt { optCiteMethod = Biblatex }))
+                 "" -- "Use biblatex cite commands in LaTeX output"
+
     , Option "" ["data-dir"]
                  (ReqArg
                   (\arg opt -> return opt { optDataDir = Just arg })
@@ -520,7 +551,7 @@ options =
                  (NoArg
                   (\_ -> do
                      prg <- getProgName
-                     hPutStrLn stdout (prg ++ " " ++ pandocVersion ++ compileInfo ++
+                     UTF8.hPutStrLn stdout (prg ++ " " ++ pandocVersion ++ compileInfo ++
                                        copyrightMessage)
                      exitWith ExitSuccess ))
                  "" -- "Print version"
@@ -529,7 +560,7 @@ options =
                  (NoArg
                   (\_ -> do
                      prg <- getProgName
-                     hPutStr stdout (usageMessage prg options)
+                     UTF8.hPutStr stdout (usageMessage prg options)
                      exitWith ExitSuccess ))
                  "" -- "Show help"
     ]
@@ -539,7 +570,7 @@ usageMessage :: String -> [OptDescr (Opt -> IO Opt)] -> String
 usageMessage programName = usageInfo
   (programName ++ " [OPTIONS] [FILES]" ++ "\nInput formats:  " ++
   (intercalate ", " $ map fst readers) ++ "\nOutput formats:  " ++
-  (intercalate ", " $ map fst writers) ++ "\nOptions:")
+  (intercalate ", " $ map fst writers ++ ["odt","epub"]) ++ "\nOptions:")
 
 -- Determine default reader based on source file extensions
 defaultReaderName :: String -> [FilePath] -> String
@@ -554,7 +585,9 @@ defaultReaderName fallback (x:xs) =
     ".ltx"      -> "latex"
     ".rst"      -> "rst"
     ".lhs"      -> "markdown+lhs"
+    ".textile"  -> "textile"
     ".native"   -> "native"
+    ".json"     -> "json"
     _           -> defaultReaderName fallback xs
 
 -- Returns True if extension of first source is .lhs
@@ -577,22 +610,26 @@ defaultWriterName x =
     ".rst"      -> "rst"
     ".s5"       -> "s5"
     ".native"   -> "native"
+    ".json"     -> "json"
     ".txt"      -> "markdown"
     ".text"     -> "markdown"
     ".md"       -> "markdown"
     ".markdown" -> "markdown"
+    ".textile"  -> "textile"
     ".lhs"      -> "markdown+lhs"
     ".texi"     -> "texinfo"
     ".texinfo"  -> "texinfo"
     ".db"       -> "docbook"
     ".odt"      -> "odt"
+    ".epub"     -> "epub"
+    ".org"      -> "org"
     ['.',y] | y `elem` ['1'..'9'] -> "man"
     _          -> "html"
 
 main :: IO ()
 main = do
 
-  rawArgs <- getArgs
+  rawArgs <- liftM (map decodeString) getArgs
   prg <- getProgName
   let compatMode = (prg == "hsmarkdown")
 
@@ -602,8 +639,8 @@ main = do
 
   unless (null errors) $
     do name <- getProgName
-       mapM_ (\e -> hPutStr stderr (name ++ ": ") >> hPutStr stderr e) errors
-       hPutStrLn stderr $ "Try " ++ name ++ " --help for more information."
+       mapM_ (\e -> UTF8.hPutStr stderr (name ++ ": ") >> UTF8.hPutStr stderr e) errors
+       UTF8.hPutStrLn stderr $ "Try " ++ name ++ " --help for more information."
        exitWith $ ExitFailure 2
 
   let defaultOpts' = if compatMode
@@ -622,47 +659,44 @@ main = do
               , optWriter            = writerName
               , optParseRaw          = parseRaw
               , optVariables         = variables
-              , optBefore            = befores
-              , optAfter             = afters
               , optTableOfContents   = toc
               , optTransforms        = transforms
               , optTemplate          = template
               , optOutputFile        = outputFile
               , optNumberSections    = numberSections
+              , optSectionDivs       = sectionDivs
               , optIncremental       = incremental
+              , optOffline           = offline
               , optXeTeX             = xetex
               , optSmart             = smart
+              , optHtml5             = html5
+              , optChapters          = chapters
               , optHTMLMathMethod    = mathMethod
               , optReferenceODT      = referenceODT
+              , optEPUBStylesheet    = epubStylesheet
+              , optEPUBMetadata      = epubMetadata
               , optDumpArgs          = dumpArgs
               , optIgnoreArgs        = ignoreArgs
               , optStrict            = strict
               , optReferenceLinks    = referenceLinks
               , optWrapText          = wrap
-              , optSanitizeHTML      = sanitize
+              , optColumns           = columns
               , optEmailObfuscation  = obfuscationMethod
               , optIdentifierPrefix  = idPrefix
               , optIndentedCodeClasses = codeBlockClasses
               , optDataDir           = mbDataDir
-#ifdef _CITEPROC
-              , optBiblioFile         = biblioFile
-              , optBiblioFormat       = biblioFormat
-              , optCslFile            = cslFile
-#endif
+              , optBibliography      = reffiles
+              , optCslFile           = cslfile
+              , optCiteMethod        = citeMethod
+              , optListings          = listings
+              , optAscii             = ascii
              } = opts
 
   when dumpArgs $
-    do hPutStrLn stdout outputFile
-       mapM_ (\arg -> hPutStrLn stdout arg) args
+    do UTF8.hPutStrLn stdout outputFile
+       mapM_ (\arg -> UTF8.hPutStrLn stdout arg) args
        exitWith ExitSuccess
 
-  -- warn about deprecated options
-  case lookup "legacy-header" variables of
-     Just _  -> hPutStrLn stderr $
-       "Warning: The -C/--custom-header is deprecated.\n" ++
-       "Please transition to using --template instead."
-     Nothing -> return ()
-
   let sources = if ignoreArgs then [] else args
 
   datadir <- case mbDataDir of
@@ -687,31 +721,25 @@ main = do
      Just r  -> return r
      Nothing -> error ("Unknown reader: " ++ readerName')
 
-  writer <- case (lookup writerName' writers) of
-     Just r  -> return r
-     Nothing -> error ("Unknown writer: " ++ writerName')
-
   templ <- getDefaultTemplate datadir writerName'
   let defaultTemplate = case templ of
                              Right t -> t
                              Left  e -> error (show e)
 
-  environment <- getEnvironment
-  let columns = case lookup "COLUMNS" environment of
-                 Just cols -> read cols
-                 Nothing   -> stateColumns defaultParserState
-
   let standalone' = standalone || isNonTextOutput writerName'
 
-#ifdef _CITEPROC
-  refs <- if null biblioFile then return [] else readBiblioFile biblioFile biblioFormat
-#endif
-
-  variables' <- if writerName' == "s5" && standalone'
-                   then do
-                     inc <- s5HeaderIncludes datadir
-                     return $ ("header-includes", inc) : variables
-                   else return variables
+  variables' <- case (writerName', standalone', offline) of
+                      ("s5", True, True) -> do
+                        inc <- s5HeaderIncludes datadir
+                        return $ ("s5includes", inc) : variables
+                      ("slidy", True, True) -> do
+                        slidyJs <- readDataFile datadir $
+                                      "slidy" </> "slidy.min.js"
+                        slidyCss <- readDataFile datadir $
+                                      "slidy" </> "slidy.min.css"
+                        return $ ("slidy-js", slidyJs) :
+                            ("slidy-css", slidyCss) : variables
+                      _ -> return variables
 
   variables'' <- case mathMethod of
                       LaTeXMathML Nothing -> do
@@ -722,62 +750,82 @@ main = do
                          return $ ("mathml-script", s) : variables'
                       _ -> return variables'
 
+  refs <- mapM (\f -> catch (readBiblioFile f) $ \e -> do
+         UTF8.hPutStrLn stderr $ "Error reading bibliography `" ++ f ++ "'"
+         UTF8.hPutStrLn stderr $ show e
+         exitWith (ExitFailure 23)) reffiles >>= \rs -> return $ concat rs
+
+  let sourceDir = if null sources
+                     then "."
+                     else takeDirectory (head sources)
+
+  let slideVariant = case writerName' of
+                           "s5"    -> S5Slides
+                           "slidy" -> SlidySlides
+                           _       -> NoSlides
+
   let startParserState =
          defaultParserState { stateParseRaw        = parseRaw,
                               stateTabStop         = tabStop,
-                              stateSanitizeHTML    = sanitize,
                               stateLiterateHaskell = "+lhs" `isSuffixOf` readerName' ||
                                                      lhsExtension sources,
                               stateStandalone      = standalone',
-#ifdef _CITEPROC
-                              stateCitations       = map citeKey refs,
-#endif
+                              stateCitations       = map refId refs,
                               stateSmart           = smart || writerName' `elem`
-                                                              ["latex", "context", "man"],
+                                                              ["latex", "context", "latex+lhs", "man"],
                               stateColumns         = columns,
                               stateStrict          = strict,
-                              stateIndentedCodeClasses = codeBlockClasses }
+                              stateIndentedCodeClasses = codeBlockClasses,
+                              stateApplyMacros     = writerName' `notElem` ["latex", "latex+lhs"] }
+
   let writerOptions = WriterOptions { writerStandalone       = standalone',
                                       writerTemplate         = if null template
                                                                   then defaultTemplate
                                                                   else template,
                                       writerVariables        = variables'',
-                                      writerIncludeBefore    = concat befores,
-                                      writerIncludeAfter     = concat afters,
+                                      writerEPUBMetadata     = epubMetadata,
                                       writerTabStop          = tabStop,
                                       writerTableOfContents  = toc &&
                                                                writerName' /= "s5",
                                       writerHTMLMathMethod   = mathMethod,
-                                      writerS5               = (writerName' == "s5"),
-                                      writerXeTeX            = xetex,
-                                      writerIgnoreNotes      = False,
+                                      writerSlideVariant     = slideVariant,
                                       writerIncremental      = incremental,
+                                      writerXeTeX            = xetex,
+                                      writerCiteMethod       = citeMethod,
+                                      writerBiblioFiles      = reffiles,
+                                      writerIgnoreNotes      = False,
                                       writerNumberSections   = numberSections,
+                                      writerSectionDivs      = sectionDivs,
                                       writerStrictMarkdown   = strict,
                                       writerReferenceLinks   = referenceLinks,
                                       writerWrapText         = wrap,
+                                      writerColumns          = columns,
                                       writerLiterateHaskell  = "+lhs" `isSuffixOf` writerName' ||
                                                                lhsExtension [outputFile],
                                       writerEmailObfuscation = if strict
                                                                   then ReferenceObfuscation
                                                                   else obfuscationMethod,
-                                      writerIdentifierPrefix = idPrefix }
+                                      writerIdentifierPrefix = idPrefix,
+                                      writerSourceDirectory  = sourceDir,
+                                      writerUserDataDir      = datadir,
+                                      writerHtml5            = html5 &&
+                                                               "html" `isPrefixOf` writerName',
+                                      writerChapters         = chapters, 
+                                      writerListings         = listings,
+                                      writerAscii            = ascii }
 
   when (isNonTextOutput writerName' && outputFile == "-") $
-    do hPutStrLn stderr ("Error:  Cannot write " ++ writerName ++ " output to stdout.\n" ++
+    do UTF8.hPutStrLn stderr ("Error:  Cannot write " ++ writerName ++ " output to stdout.\n" ++
                                "Specify an output file using the -o option.")
        exitWith $ ExitFailure 5
 
-  let sourceDirRelative = if null sources
-                             then ""
-                             else takeDirectory (head sources)
-
   let readSources [] = mapM readSource ["-"]
       readSources srcs = mapM readSource srcs
-      readSource "-" = getContents
+      readSource "-" = UTF8.getContents
       readSource src = case parseURI src of
-                            Just u  -> readURI u
-                            Nothing -> readFile src
+                            Just u | uriScheme u `elem` ["http:","https:"] ->
+                                       readURI u
+                            _       -> UTF8.readFile src
       readURI uri = simpleHTTP (mkRequest GET uri) >>= getResponseBody >>=
                       return . toString  -- treat all as UTF8
 
@@ -785,19 +833,39 @@ main = do
 
   doc <- fmap (reader startParserState . convertTabs . intercalate "\n") (readSources sources)
 
-  let doc' = foldr ($) doc transforms
+  let doc0 = foldr ($) doc transforms
 
-  doc'' <- do
-#ifdef _CITEPROC
-          processBiblio cslFile refs doc'
-#else
-          return doc'
-#endif
+  doc1 <- if writerName' == "rtf"
+             then bottomUpM rtfEmbedImage doc0
+             else return doc0
 
-  let writerOutput = writer writerOptions doc'' ++ "\n"
+  doc2 <- do
+          if citeMethod == Citeproc && not (null refs)
+             then do
+                csldir <- getAppUserDataDirectory "csl"
+                cslfile' <- if null cslfile
+                               then findDataFile datadir "default.csl"
+                               else do
+                                  ex <- doesFileExist cslfile
+                                  if ex
+                                     then return cslfile
+                                     else findDataFile datadir $
+                                            replaceDirectory
+                                            (replaceExtension cslfile "csl")
+                                            csldir
+                processBiblio cslfile' refs doc1
+             else return doc1
 
-  case writerName' of
-       "odt"   -> saveOpenDocumentAsODT datadir outputFile sourceDirRelative referenceODT writerOutput
-       _       -> if outputFile == "-"
-                     then putStr writerOutput
-                     else writeFile outputFile writerOutput
+  case lookup writerName' writers of
+        Nothing | writerName' == "epub" ->
+           writeEPUB epubStylesheet writerOptions doc2
+           >>= B.writeFile (encodeString outputFile)
+        Nothing | writerName' == "odt"  ->
+           writeODT referenceODT writerOptions doc2
+           >>= B.writeFile (encodeString outputFile)
+        Just r  -> writerFn outputFile result
+            where writerFn "-" = UTF8.putStr
+                  writerFn f   = UTF8.writeFile f
+                  result       = r writerOptions doc2 ++
+                                 ['\n' | not standalone']
+        Nothing -> error $ "Unknown writer: " ++ writerName'
